# -*- coding: utf-8 -*-
import locale
from calendar import monthrange
from random import choice
from string import printable
from platform import system
import logging
import json
import datetime as dt
from dateutil.relativedelta import relativedelta
from dateutil.rrule import rrule, \
    SU, MO, TU, WE, TH, FR, SA, \
    DAILY, WEEKLY, MONTHLY
from dateutil.parser import parser
from enum import IntEnum

locale.setlocale(locale.LC_ALL, '')    # set your locale

logger = logging.getLogger(__name__)


a_file = '/home/spiros/.config/pyradio/data/schedule.json'


def is_date_before(a_date, b_date):
    if (a_date - b_date).days < 0:
        return True
    return False

def is_date_after(a_date, b_date):
    return not is_date_before(a_date, b_date)

def datetime_to_my_time(a_date, shorten=True):
    t = a_date.ctime().split()
    now = dt.datetime.now()
    if shorten:
        if a_date.date() == now.date():
            return 'At ' + t[-2]
        # now = now.replace(day=now.day+1)
        now = now + dt.timedelta(days=1)
        if a_date.date() == now.date():
            return 'Tomorrow, at ' + t[-2]

    if dt.datetime.now().year == a_date.year:
        return t[0] + ', ' + t[1] + ' ' + t[2] + ', ' + t[-2]
    return t[0] + ', ' + t[1] + ' ' + t[2] + ' ' + t[-1] +', ' + t[-2]

def random_string(length=10):
    return ''.join(choice(printable) for i in range(length))

def format_date_to_iso8851(a_date=None):
    ''' format a datetime.date in ISO-8851 format
    '''
    if a_date is None:
        a_date = dt.datetime.now()
    days = ('Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun')
    months = ('Jan', 'Feb', 'Mar', 'Apr', 'May', 'Jun',
              'Jul', 'Aug', 'Sep', 'Oct', 'Nov', 'Dec')

    cur_day = days[a_date.weekday()]
    cur_month = months[a_date.month-1]
    return cur_day + ', ' + str(a_date.day) + ' ' + \
        cur_month + ' ' + str(a_date.year) + \
        a_date.strftime(' %H:%M:%S')

class PyRadioScheduleItemType(IntEnum):
    TYPE_START_END = 0
    TYPE_START = 1
    TYPE_END = 2

    @property
    def items(cls):
        return tuple(member.value for member in cls)

    @classmethod
    def to_string(cls, a_type):
        try:
            return cls(a_type).name
        except ValueError:
            return 'UNKNOWN'

class PyRadioScheduleTimeType(IntEnum):

    TIME_ABSOLUTE = 0
    TIME_RELATIVE = 1

    @classmethod
    def items(cls):
        return tuple(member.value for member in cls)

    @classmethod
    def to_string(cls, a_type):
        try:
            return cls(a_type).name
        except ValueError:
            return 'UNKNOWN'


class PyRadioScheduleList():

    def __init__(self, a_file, a_list=None):
        '''
            a_file : file to read
            a_list : list of dicts of (as output of self._read_list)
        '''
        self._list = []              # initial list of dicts
        self._schedule_list = []     # lisr of PyRadioScheduleItem
        self._sorted = []            # list (sorted) of tasks (dicts)
        self._info = False
        self._json_file = a_file
        self._a_list = a_list
        self._read_list()

    def _read_list(self):
        ''' read json file to self._list '''
        if self._a_list:
            logger.error(f'\n\nself._list\n{self._a_list}\n\n')
            self._list = self._a_list
            for _i in range(len(self._list)):
                if self._list[-1]['token'] == '':
                    self._list[-1]['token'] = random_string()
        else:
            try:
                with open(self._json_file, 'r', encoding='utf-8') as f:
                    self._list = json.load(f)
            except:
                self._list = []

    def _list_to_schedule_items(self):
        self._schedule_list = []
        for i, _ in enumerate(self._list):
            self._schedule_list.append(
                    PyRadioScheduleItem(self._list[i])
                    )
        for i in range(len(self._schedule_list)-1, -1, -1):
            ac = self._schedule_list[i].get_active_item()
            now = dt.datetime.now()
            if is_date_before(ac[1], now) and \
                    is_date_before(ac[2], now) and \
                    ac[7] is None:
                self._schedule_list.pop(i)
        # logger.error('\n\nself._schedule_list\n{}'.format(self._schedule_list))

    def get_list_of_active_items(self):
        if self._schedule_list == []:
            # logger.error('len(self._list) = {}'.format(len(self._list)))
            for i, _ in enumerate(self._list):
                self._schedule_list.append(
                        PyRadioScheduleItem(self._list[i])
                        )
        return [x.get_active_item() for x in self._schedule_list]

    def get_info_of_items(self):
        ''' Returns a list of items comming from self._schedule_list
            (which is a list of PyRadioScheduleItem), to be presented
            to the user
        '''
        out = []
        for i, n in enumerate(self._schedule_list):
            x = n.get_active_item()
            out.append(f'Item {i + 1}')
            if x[2] == 0 or x[2] == 1:
                out.append('  Start playback on: ' + datetime_to_my_time(x[0]))
            if x[2] == 0 or x[2] == 2:
                out.append('   Stop playback on: ' + datetime_to_my_time(x[1]))
            if x[2] == 0 or x[2] == 1:
                out.append('  From playlist: ' + x[-3])
                out.append('        Station: ' + x[-2])
                out.append('  Player: ' + x[3])
                if x[-4]:
                    out.append('  Repeat every: ' + x[-4])
                if x[-6] == 0:
                    tmp = '  Recording is off,'
                elif x[-6] == 1:
                    tmp = '  Recording is on,'
                elif x[-6] == 2:
                    tmp = '  Recording is on (silent),'
                if x[-5] == 0:
                    tmp += ' with no buffering'
                else:
                    tmp += ' with buffering'
                out.append(tmp)
        return out

    def get_list_of_repeating_dates(self, count):
        if self._schedule_list == []:
            self._list_to_schedule_items()
        rep = self.get_repeating_dates(self._schedule_list[0].get_active_item(), count=count)
        return [(datetime_to_my_time(x[1]), datetime_to_my_time(x[2])) for x in rep]

    def get_repeating_dates(self, in_date, count=None):
        days = {
            'Sunday': SU,
            'Monday': MO,
            'Tuesday': TU,
            'Wednesday': WE,
            'Thursday': TH,
            'Friday': FR,
            'Saturday': SA
        }

        rule = in_date[-4]
        start_dt = in_date[1]
        end_dt = in_date[2]
        now = dt.datetime.now()
        now_with_correct_time = now.replace(
            hour=start_dt.hour,
            minute=start_dt.minute,
            second=start_dt.second,
            microsecond=0
        )
        now_plus_seven_days = now_with_correct_time + dt.timedelta(days=7)

        ll = []

        if count is None:
            if rule == 'day':
                # Keep a few close daily occurrences
                ll = list(rrule(DAILY, count=12, dtstart=start_dt))
                ll = [x for x in ll if now < x <= now_plus_seven_days]

            elif rule == 'week':
                # See a bit into the future and keep the closer ones
                ll = list(rrule(WEEKLY, dtstart=start_dt, until=now + dt.timedelta(days=14)))
                ll = [x for x in ll if now < x <= now_plus_seven_days]

            elif rule == 'month':
                # See about 2 months forward and keep anything within the 7 day preview window
                ll = list(rrule(MONTHLY, dtstart=start_dt, until=now + dt.timedelta(days=65)))
                ll = [x for x in ll if now < x <= now_plus_seven_days]

            elif rule in days:
                # For a specifies weekday keep a few future occurrences
                ll = list(
                    rrule(
                        WEEKLY,
                        count=3,
                        wkst=SU,
                        byweekday=(days[rule],),
                        dtstart=start_dt
                    )
                )
                ll = [x for x in ll if now < x <= now_plus_seven_days]

            else:
                return []

        else:
            if rule == 'day':
                ll = list(rrule(DAILY, count=count, dtstart=start_dt))

            elif rule == 'week':
                ll = list(rrule(WEEKLY, count=count, dtstart=start_dt))

            elif rule == 'month':
                ll = list(rrule(MONTHLY, count=count, dtstart=start_dt))

            elif rule in days:
                ll = list(
                    rrule(
                        WEEKLY,
                        count=count,
                        wkst=SU,
                        byweekday=(days[rule],),
                        dtstart=start_dt
                    )
                )

            else:
                return []

        out = []
        if ll:
            duration = end_dt - start_dt
            for x in ll:
                item = list(in_date)
                item[1] = x
                item[2] = x + duration
                out.append(item)

        return out

    def get_info_of_tasks(self, to_dict=False):
        ''' Returns a list of tasks coming from self._sorted
            to be presented to the user
        '''
        self.get_list_of_tasks()
        out = []
        if to_dict:
            links_found = 0
            tick = ' X ' if system().lower().startswith('win') else ' ✔ '
            for i, n in enumerate(self._sorted):
                if 'player' in n:
                    tmp = {
                            'id': str(i+1),
                            'name': n['name'],
                            'link': '',
                            'start': datetime_to_my_time(n['date']),
                            'stop': '',
                            'player': n['player'],
                            'playlist': n['playlist'],
                            'station': n['station'],
                            'recording': '' if n['recording'] == 0  else tick,
                            'buffering': '' if n['buffering'] == 0 else tick
                    }
                    if n['recording'] == 2:
                        tmp['recording'] = 'sil'
                else:
                    if out:
                        link = self._get_linked_task(n['link'], True)
                        if link:
                            out[int(link)-links_found-1]['stop'] = datetime_to_my_time(n['date'])
                            links_found += 1
                            continue
                    tmp = {
                            'id': str(i+1),
                            'link': self._get_linked_task(n['link'], to_dict),
                            'start': '',
                            'stop': datetime_to_my_time(n['date']),
                            'player': '',
                            'playlist': '',
                            'station': '',
                            'recording': '',
                            'buffering': ''
                    }
                out.append(tmp)
            for i, _ in enumerate(out):
                if out[i]['stop'][:13] == out[i]['start'][:13]:
                    out[i]['stop'] = out[i]['stop'][13:]
        else:
            for i, n in enumerate(self._sorted):
                if 'player' in n:
                    out.append(f'Task: {i + 1}')
                    out.append('    Name: ' + n['name'])
                    out.append('    Start playback on: ' + datetime_to_my_time(n['date']))
                    out.append('    From playlist: ' + n['playlist'])
                    out.append('          Station: ' + n['station'])
                    if n['recording'] == 0:
                        tmp = '    Recording is off,'
                    elif n['recording'] == 1:
                        tmp = '    Recording is on,'
                    elif n['recording'] == 2:
                        tmp = '    Recording is on (silent),'
                    if n['buffering'] == 0:
                        tmp += ' with no buffering'
                    else:
                        tmp += ' with buffering'
                    out.append(tmp)
                else:
                    out.append('').format(i + 1, self._get_linked_task(n['link']))
                    out.append('     Stop playback on: ' + datetime_to_my_time(n['date']))
        return out

    def _get_linked_task(self, a_link, to_dict=False):
        for i, n in enumerate(self._sorted):
            if a_link == n['link']:
                if to_dict:
                    return str(i + 1)
                return f'  ->   linked to Task {i + 1}'
        return ''

    def get_list_of_tasks(self):
        self._sorted = []
        if self._schedule_list == []:
            self._list_to_schedule_items()

        # for n in self._schedule_list:
        #     logger.error(n)
        # logger.error('\n\n')

        a_list= []
        for i, _ in enumerate(self._schedule_list):
            if self._schedule_list[i].repeat is None:
                a_list.append(
                    self._schedule_list[i].get_active_item()
                )
            else:
                a_list.extend(
                    self.get_repeating_dates(
                        self._schedule_list[i].get_active_item(),
                    )
                )

            # logger.info('\n\na_list')
            # for n in a_list:
            #     logger.error('\n{}'.format(n))
            # logger.info('\n\n')

        if a_list:
            a_list.sort(key=lambda x: x[0])
            for n in a_list:
                self._sorted.append({
                    'name': n[0],
                    'date': n[1],
                    'player': n[4],
                    'recording': n[5],
                    'buffering': n[6],
                    'playlist': n[8],
                    'station': n[9],
                    'token': n[-1] if n[-1] else random_string(),
                    'link': random_string() if n[3] == 0 else ''
                })
                if n[3] == 0:
                    self._sorted.append({
                        'date': n[2],
                        'token': self._sorted[-1]['token'],
                        'link': self._sorted[-1]['link']
                    })

        # logger.error('\n\nself._sorted')
        # for n in self._sorted:
        #     logger.error(n)
        # logger.error('\n\n')

    def item(self, index):
        if index < len(self._list):
            return self._list[index]
        return None

    def count(self):
        return len(self._schedule_list)

class PyRadioScheduleItem():
    '''
    Provide a Schedule Item

    Items Format:
        type:
            "B" "S" or "E"
                "B": Item has both a start and an end time
                "S": Item has only a start time
                "E": Item has only an end time
                     Start time is the time the Item was created

        start_date
            A PyRadioDate: [YYYY, M, D]
        start_time
            A PyRadioTime: [X, X, X, T]

        start type
            "A" or "I"
                "A": Date and Time is absolute
                "I": Date and Time is relative
                     In this case, the field becomes "IXX:XX:XX"
                     and the start date is the creation date
            This gives a hint for the presenting window

        end_date
            A PyRadioDate: [YYYY, M, D]
        end_time
            A PyRadioTime: [X, X, X, T]

        end type
            "A" or "I"
                "A": Date and Time is absolute
                "I": Date and Time is relative
                     In this case, the field becomes "IXX:XX:XX"
                     and the end date is the creation date
            This gives a hint for the presenting window

        playlist
            Playlist title

        station
            station name

    '''

    def __init__(self, item=None):
        self._item = None
        if item:
            if isinstance(item, str):
                self._item = json.loads(item)
            elif isinstance(item, dict):
                self._item = item

        if self._item is None:
            self._item = self.default_item

    def __str__(self):
        out = ['Current _item:']
        for n in self._item.keys():
            s = "  " + "'{}':  ".format(n).rjust(20)
            if n == 'type':
                out.append(s + "{} (".format(self._item[n]) + \
                        PyRadioScheduleItemType.to_string(self._item['type']) +
                           ')  -> '  + str(type(self.item[n]))
                )
            elif n in ('start_type', 'end_type'):
                out.append(s + "{} (".format(self._item[n]) + \
                        PyRadioScheduleTimeType.to_string(self._item[n]) \
                        + ')  -> ' + str(type(self._item[n]))
                )
            elif n in ('start_time', 'end_time', 'start_duration', 'end_duration'):
                t = self._item[n][-1]
                x = self._item[n][:-1]
                out.append(s + str(x)[:-1] + \
                        ', {0} ({1})]  -> {2}'.format(t, PyRadioTime.to_string(t), type(self._item[n])))
            else:
                out.append(s + str(self._item[n]) + '  -> ' + str(type(self._item[n])))
        return '\n'.join(out)

    @property
    def active_item(self):
        '''return the item after calculating all relative values'''
        name, st, en, it_type, pl, rec, buf, rep, playlist, station, token = self.get_active_item()
        if token == '':
            token = random_string()
        out = {
            'name': name,
            'type': it_type,
            'start_type': 0,
            'start_date': [st.year, st.month, st.day],
            'start_time': [st.hour, st.minute, st.second, 0],
            'start_duration': [0, 0, 0, 0],
            'end_type': 0,
            'end_date': [en.year, en.month, en.day],
            'end_time': [en.hour, en.minute, en.second, 0],
            'end_duration': [0, 0, 0, 0],
            'playlist': playlist,
            'station': station,
            'player': pl,
            'recording': rec,
            'buffering': buf,
            'repeat': rep,
            'token': token
        }
        return out

    @property
    def item(self):
        return self._item

    @item.setter
    def item(self, val):
        if isinstance(val, str):
            try:
                self._item = json.loads(val)
            except json.JSONDecodeError as e:
                raise ValueError('JSON string not supported') from e
        elif isinstance(val, dict):
            self._item = val

        for n in self.default_item:
            if n not in self._item:
                self.item = self.default_item
                raise ValueError('Item is missing keys')

        '''make sure duration times are of NO_AM_PM_FORMAT'''
        self._item['start_duration'][-1] = 0
        self._item['end_duration'][-1] = 0

    @property
    def default_item(self):
        n_date, n_time = self._get_today_plus_one_hour()
        t_date, t_time = self._get_today()
        return {
            'name': 'Default schedule entry',
            'type': 2, # TYPE_START_END, TYPE_START, TYPE_END
            'start_type': 0, # TIME_ABSOLUTE, TIME_RELATIVE
            'start_date':  t_date, # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'start_time': t_time, # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'start_duration': [0, 0, 0, 0], # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'end_type': 1, # TIME_ABSOLUTE, TIME_RELATIVE
            'end_date': n_date,
            'end_time': n_time, # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'end_duration': [1, 0, 0, 0], # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
            'player': '',
            'recording': 0,
            'buffering': 0,
            'repeat': None,
            'playlist': None,
            'station': None,
            'token': random_string()
        }

    @property
    def type(self):
        return self._item['type']

    @type.setter
    def type(self, val):
        try:
            if val in PyRadioScheduleItemType.items():
                self._item['type'] = val
                return
        except:
            pass
        raise ValueError('Invalid item type')

    @property
    def start_type(self):
        return self._item['start_type']

    @start_type.setter
    def start_type(self, val):
        self._item['start_type'] = val

    @property
    def start_date(self):
        return self._item['start_date']

    @start_date.setter
    def start_date(self, val):
        self._item['start_date'] = val

    @property
    def start_time(self):
        return self._item['start_time']

    @start_time.setter
    def start_time(self, val):
        self._item['start_time'] = val

    @property
    def start_duration(self):
        return self._item['start_duration']

    @start_duration.setter
    def start_duration(self, val):
        self._item['start_duration'] = val

    @property
    def end_type(self):
        return self._item['end_type']

    @end_type.setter
    def end_type(self, val):
        self._item['end_type'] = val

    @property
    def end_date(self):
        return self._item['end_date']

    @end_date.setter
    def end_date(self, val):
        self._item['end_date'] = val

    @property
    def end_time(self):
        return self._item['end_time']

    @end_time.setter
    def end_time(self, val):
        self._item['end_time'] = val

    @property
    def end_duration(self):
        return self._item['end_duration']

    @end_duration.setter
    def end_duration(self, val):
        self._item['end_duration'] = val

    @property
    def recording(self):
        return self._item['recording']

    @recording.setter
    def recording(self, val):
        self._item['recording'] = val

    @property
    def buffering(self):
        return self._item['buffering']

    @buffering.setter
    def buffering(self, val):
        self._item['buffering'] = val

    @property
    def repeat(self):
        return self._item['repeat']

    @repeat.setter
    def repeat(self, val):
        self._item['repeat'] = val

    @property
    def playlist(self):
        return self._item['playlist']

    @playlist.setter
    def playlist(self, val):
        self._item['playlist'] = val

    @property
    def station(self):
        return self._item['station']

    @station.setter
    def station(self, val):
        self._item['station'] = val

    @property
    def string(self):
        return json.dumps(self._item)

    def _check_start_end_type(self, val):
        try:
            if val in (
                PyRadioScheduleTimeType.TIME_ABSOLUTE,
                PyRadioScheduleTimeType.TIME_RELATIVE,
            ):
                self._item['start_type'] = val
                return True
        except:
            pass
        return False

    def get_active_item(self):
        '''
        return a tuple of datetimes representing
        (
            name
            starting date-time,
            ending date-time,
            type (start/stop/both)
            player to use
            recording (0 if TYPE_END)
            buffering (0 if TYPE_END
            repeat (None if TYPE_END)
            playlist (None if TYPE_END)
            station (None if TYPE_END)
            token
        )
        '''
        today = dt.datetime.now().replace(microsecond=0)
        if self._item['type'] in (
                PyRadioScheduleItemType.TYPE_START_END,
                PyRadioScheduleItemType.TYPE_START
        ):
            if self._item['start_type'] == PyRadioScheduleTimeType.TIME_ABSOLUTE:
                start_date = dt.datetime(
                    year=self._item['start_date'][0],
                    month=self._item['start_date'][1],
                    day=self._item['start_date'][2],
                ) + PyRadioTime.pyradio_time_to_timedelta(
                    self._item['start_time']
                )
            else:
                start_date = today + PyRadioTime.pyradio_time_to_timedelta(
                    self._item['start_duration']
                )


        if self._item['type'] in (
                PyRadioScheduleItemType.TYPE_START_END,
                PyRadioScheduleItemType.TYPE_END
        ):
            if self._item['end_type'] == PyRadioScheduleTimeType.TIME_ABSOLUTE:
                end_date = dt.datetime(
                    self._item['end_date'][0],
                    self._item['end_date'][1],
                    self._item['end_date'][2]
                ) + dt.timedelta(
                    hours=self._item['end_time'][0],
                    minutes=self._item['end_time'][1],
                    seconds=self._item['end_time'][2],
                )
            else:
                if self._item['type'] == PyRadioScheduleItemType.TYPE_END:
                    use_date = today
                else:
                    use_date = start_date
                end_date = use_date + dt.timedelta(
                    hours=self._item['end_duration'][0],
                    minutes=self._item['end_duration'][1],
                    seconds=self._item['end_duration'][2]
                )

        if self._item['type'] == PyRadioScheduleItemType.TYPE_START:
            end_date = start_date
        elif self._item['type'] == PyRadioScheduleItemType.TYPE_END:
            start_date = end_date
        rep = self._item['repeat']
        if rep is not None:
            rep = rep.strip()
        return (self._item['name'],
                start_date, end_date,
                self._item['type'],
                self._item['player'],
                self._item['recording'],
                self._item['buffering'],
                rep,
                self._item['playlist'] if self._item['type'] != PyRadioScheduleItemType.TYPE_END else None,
                self._item['station'] if self._item['type'] != PyRadioScheduleItemType.TYPE_END else None,
                self._item['token'],
                )

    def _get_today(self):
        ''' get today as pyradio date and time '''
        today = dt.datetime.now()
        return [today.year, today.month, today.day], \
                [today.hour, today.minute, 0, 0]

    def _get_today_plus_one_hour(self):
        today = dt.datetime.now() + relativedelta(hours=+1)
        return [today.year, today.month, today.day], \
            [today.hour, today.minute, 0, 0]


class PyRadioTime():
    ''' A class to provide PyRadio time and date

        PyRadio time is a tuple
        Its format is ALWAYS in 24-hour notation
        and contains
            hour, minutes, seconds, type

            type shows how the time was inserted
                type is 0: 24-hour format
                        1: AM
                        2: PM

        Parameters
        ==========
        date
            Date string
                format is YYYY-MM-DD
            if None, date will not be set
        time
            Time string
                format is XX:XX [AM/PM]
                          XX:XX:XX [AM/PM]
    '''

    NO_AM_PM_FORMAT = 0
    AM_FORMAT = 1
    PM_FORMAT = 2

    def __init__( self):
        '''
        '''
        self.date = None        # datetime.date
        self.time = None        # pyradio time

    def __str__(self):
        if self.date is None:
            self.date = dt.date.today()
        if self.time is None:
            self.time = PyRadioTime.string_to_pyradio_time(dt.datetime.now().strftime('%H:%M:%S'))
        return self.date.strftime('%Y-%m-%d') + ' ' + PyRadioTime.pyradio_time_to_string(self.time)

    @classmethod
    def to_string(cls, a_time_format):
        if a_time_format == 1:
            return 'AM_FORMAT'
        if a_time_format == 2:
            return 'PM_FORMAT'
        return 'NO_AM_PM_FORMAT'

    def set_date_and_time(self, a_date_time_string):
        sp = a_date_time_string.split(' ')
        self.set_date(sp[0])
        if len(sp) == 3:
            self.set_time(sp[1] + ' ' + sp[2])
        else:
            self.set_time(sp[1])

    def set_date(self, a_date_string):
        if a_date_string:
            self.date = dt.date.fromisoformat(a_date_string)
        else:
            self.date = dt.date.today()

    def set_time(self, a_time_string):
        valid = True
        is_12_format = self.NO_AM_PM_FORMAT
        if a_time_string is None:
            valid = False
        else:
            if a_time_string.lower().endswith(' am'):
                is_12_format = self.AM_FORMAT
                my_string = a_time_string[:-3]
            elif a_time_string.lower().endswith(' pm'):
                is_12_format = self.PM_FORMAT
                my_string = a_time_string[:-3]
            else:
                my_string = a_time_string
            sp = my_string.split(':')
            if len(sp) > 3:
                valid = False
            if len(sp) < 3:
                sp.append('00')
            try:
                hour = int(sp[0])
                minute = int(sp[1])
                seconds = int(sp[2])
            except ValueError:
                valid = False

        hour = 0
        minute = 0
        seconds  = 0
        if valid:
            if is_12_format == self.PM_FORMAT:
                ''' pm '''
                hour += 12
                if hour == 24:
                    hour =0
                elif hour > 24:
                    valid = False
            elif is_12_format == self.AM_FORMAT:
                ''' am '''
                if hour > 12:
                    valid = False

        if not valid:
            a_date_time = dt.datetime.now()
            hour = a_date_time.hour
            minute = a_date_time.minute
            seconds = a_date_time.second

        self.time = [hour, minute, seconds, is_12_format]

    @classmethod
    def number_of_days_in_month(self, year, month):
        return monthrange(int(year), int(month))

    @classmethod
    def string_to_pyradio_time(cls, a_string):
        ''' convert string to hour,minutes,seconds

            Input can be
                XX:XX [AM/PM]
                XX:XX:XX [AM/PM]

            Returns hour, minutes, seconds, type
                output format is ALWAYS in 24-hour
                type shows how the time was inserted
                    type is 0: 24-hour format
                            1: AM
                            2: PM

            if time is invalid or None, return current time
                in 24-hour format
        '''

        valid = True
        is_12_format = PyRadioTime.NO_AM_PM_FORMAT
        if a_string is None:
            valid = False
        else:
            if a_string.lower().endswith(' am'):
                is_12_format = PyRadioTime.AM_FORMAT
                my_string = a_string[:-3]
            elif a_string.lower().endswith(' pm'):
                is_12_format = PyRadioTime.PM_FORMAT
                my_string = a_string[:-3]
            else:
                my_string = a_string
            sp = my_string.split(':')
            if len(sp) > 3:
                valid = False
            if len(sp) < 3:
                sp.append('00')
            try:
                hour = int(sp[0])
                minute = int(sp[1])
                seconds = int(sp[2])
            except ValueError:
                valid = False

        hour = 0
        minute = 0
        seconds = 0
        if valid:
            if is_12_format == PyRadioTime.PM_FORMAT:
                ''' pm '''
                hour += 12
                if hour == 24:
                    hour = 0
                elif hour > 24:
                    valid = False
            elif is_12_format == PyRadioTime.AM_FORMAT:
                ''' am '''
                if hour > 12:
                    valid = False

        if not valid:
            a_date_time = dt.datetime.now()
            hour = a_date_time.hour
            minute = a_date_time.minute
            seconds = a_date_time.second

        return hour, minute, seconds, is_12_format

    @classmethod
    def pyradio_time_to_string(cls, a_time):
        a_date = dt.datetime(2022, 1, 1) + dt.timedelta(
            hours=a_time[0],
            minutes=a_time[1],
            seconds=a_time[2]
        )

        if a_time[-1] == 0:
            ''' 24-hour format '''
            return a_date.strftime('%H:%M:%S')

        if a_time[-1] == 2 and a_date.hour > 12:
            a_date = a_date.replace(hour=a_date.hour - 12)
        return '{0} {1}'.format(
            a_date.strftime('%H:%M:%S'),
            'PM' if a_time[-1] == PyRadioTime.PM_FORMAT else 'AM'
        )

    @classmethod
    def to_24_format(cls, a_time):
        if a_time[-1] == PyRadioTime.NO_AM_PM_FORMAT:
            return a_time

        if a_time[-1] == PyRadioTime.PM_FORMAT:
            ''' PM '''
            out = list(a_time)
            out[-1] = 0
            return tuple(out)

        out = list(a_time)
        out[-1] = 0
        out[0] += 12
        return tuple(out)

    @classmethod
    def pyradio_time_to_timedelta(cls, a_time):
        if a_time[-1] == PyRadioTime.NO_AM_PM_FORMAT:
            return dt.timedelta(
                hours=a_time[0],
                minutes=a_time[1],
                seconds=a_time[2]
            )

        if a_time[-1] == PyRadioTime.AM_FORMAT:
            return dt.timedelta(
                hours=a_time[0] if a_time[0] < 12 else a_time[0] - 12,
                minutes=a_time[1],
                seconds=a_time[2]
            )

        if a_time[-1] == PyRadioTime.PM_FORMAT:
            return dt.timedelta(
                hours=a_time[0] + 12,
                minutes=a_time[1],
                seconds=a_time[2]
            )

    def schedule_datetime(self):
        ''' return final datetime for this object
            use this for further caclulations
        '''
        return self.pyradio_time_to_datetime(t_time=self.time)

    def pyradio_time_to_datetime(self, t_date=None, t_time=None):
        if t_time is None:
            a_time = self.time
        else:
            a_time = t_time
        if t_date is None:
            a_date = dt.datetime.now()
        else:
            a_date = t_date
        b_date= a_date.replace(
            hour = a_time[0],
            minute = a_time[1],
            second = a_time[2],
            microsecond = 0
        )
        return b_date

    @classmethod
    def seconds_to_sting(cls, time_in_seconds):
        return PyRadioTime.delta_to_sting(time_in_seconds, is_seconds=True)

    @classmethod
    def delta_to_sting(cls, time_delta, is_seconds=False):
        if is_seconds:
            s = time_delta
        else:
            s = time_delta.total_seconds()
        hours, remainder = divmod(s, 3600)
        minutes, seconds = divmod(remainder, 60)
        if hours > 0:
            out = '{:02}:{:02}:{:02}'.format(int(hours), int(minutes), int(seconds))
        else:
            out = '{:02}:{:02}'.format(int(minutes), int(seconds))
        return out

    @classmethod
    def pyradio_time_to_seconds(csl, t_time):
        ''' return number od seconds in a pyradio time'''
        return 3600 * t_time[0] + 60 * t_time[1] + t_time[2]

    @classmethod
    def pyradio_time_diference_in_seconds(cls, s_time, e_time):
        ''' return the difference of two pyradio times
            in seconds

            Paremeters
            ==========
            s_time start time
            e_time end time

            if s_time > e_time, this is a negative number
        '''
        return PyRadioTime.pyradio_time_to_seconds(e_time) - \
            PyRadioTime.pyradio_time_to_seconds(s_time)



if __name__ == '__main__':
    # print(format_date_to_iso8851(dt.datetime.now()))

    # print('\n\n============')
    # an_item = {
    #     'type': PyRadioScheduleItemType.TYPE_START_END,
    #     'start_type': 0,                            # TIME_ABSOLUTE, TYPE_RELATIVE
    #     'start_date':  [2022, 10, 15],
    #     'start_time': [8, 15, 12, 2],              # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
    #     'start_duration': [1, 0, 0, 0],
    #     'end_type': 1,                              # TIME_ABSOLUTE, TYPE_RELATIVE
    #     'end_date': [2023, 1, 1],
    #     'end_time': [3, 12, 2, 1],                  # NO_AM_PM_FORMAT, AM_FORMAT, PM_FORMAT
    #     'end_duration': [5, 15, 11, 0],
    #     'player': 'mpv',
    #     'recording': 0,
    #     'buffering': 0,
    #     'repeat': None,
    #     'playlist': 'myplaylist',
    #     'station': 'mystation',
    #     'token': random_string()
    # }

    # for n,k in an_item.items():
    #     print(n, ":", k)
    # print('\n\n============')

    # b = PyRadioScheduleItem(an_item)
    # print('b =', b)
    # u = b.get_active_item()
    # print('\n\n')
    # print(b.get_active_item())
    # print('start_date:', str(u[0]))
    # print('  end_date:', str(u[1]))
    # print('      type:', PyRadioScheduleItemType.to_string(u[2]))
    # print('  playlist:', u[3])
    # print('   station:', u[4])

    # print('\n\n{}'.format(b.string))


    my_list = [{
        'name': 'A test',
        'type': 0,
        'start_type': 0,
        'start_date': [2026, 11, 19],
        'start_time': (18, 23, 0, 0),
        'start_duration': (1, 32, 15, 0),
        'end_type': 0,
        'end_date': [2026, 11, 19],
        'end_time': (20, 1, 2, 0),
        'end_duration': (3, 15, 2, 0),
        'player': 'mpv',
        'recording': 0,
        'buffering': 0,
        'repeat': 'month',
        'playlist': 'reversed',
        'station': 'The UK 1940s Radio Station  1920s 1930s 1940s',
        'token': 'LhhO\x0c@>BET'
    }]
    x = PyRadioScheduleList(a_file='', a_list=my_list)
    # x = PyRadioScheduleList(a_file='/home/spiros/.config/pyradio/data/schedule.json')
    out = x.get_list_of_active_items()
    print(type(out))
    print(f'{out = }')


    print('\n\nper item')
    for n in out:
        print(n)
    print('\n\n')


    if x._schedule_list == []:
        x._list_to_schedule_items()

    print(type(x))
    y = PyRadioScheduleItem(x)
    print('active_item')
    print(y.get_active_item())


    print('Repeating dates')
    rep = x.get_repeating_dates(x._schedule_list[0].get_active_item(), count=6)
    print('\n\n')
    for n in rep:
        print(n)
    # x.get_list_of_tasks()
    # print('\n\n')
    # for n in x._sorted:
    #     print(n)
    print('\n\n')

    for n in rep:
        print('n[1] =', n[1])
        print('type(n[1]) =', type(n[1]))
        print(datetime_to_my_time(n[1]))
        print(datetime_to_my_time(n[2]))
        print('')

    r = x.get_list_of_repeating_dates(6)
    print('\n\n')
    for n in r:
        print(n)

    print('\n\n\n')

    t = PyRadioTime()
    print(t)

    t.time = [11, 15, 34, 1]
    print(t)

    print(f'\n\n{datetime_to_my_time(dt.datetime.now())}')

    print(f'\n\n{datetime_to_my_time(dt.datetime.now() + dt.timedelta(days=1))}')




# -*- coding: utf-8 -*-
import locale
import sys
import subprocess
from os.path import join, exists, isdir, basename, dirname
from os import environ, makedirs, listdir, replace, remove, sep, getenv, chdir, startfile, path
from time import sleep
import site
from shutil import rmtree
from msvcrt import getwch
from urllib.request import urlretrieve
import glob
import threading
import requests
try:
    # Python ≥ 3.9
    from importlib.resources import files, as_file
    from importlib.resources.abc import Traversable
except ImportError:
    # Python 3.7 & 3.8 (backport)
    from importlib_resources import files, as_file
    from importlib_resources.abc import Traversable

locale.setlocale(locale.LC_ALL, "")

HAVE_PYUNPACK = True
try:
    from pyunpack import Archive
except ModuleNotFoundError:
    HAVE_PYUNPACK = False

''' This is also to be able to execute it manually'''
try:
    from .player import find_mpv_on_windows, find_mplayer_on_windows, find_vlc_on_windows
except ImportError:
    from player import find_mpv_on_windows, find_mplayer_on_windows, find_vlc_on_windows

purl = (
    'https://sourceforge.net/projects/mpv-player-windows/files/64bit',
    'https://sourceforge.net/projects/mplayerwin/files/MPlayer-MEncoder'
)

'''
    to get the latest v3 version use this url
        https://sourceforge.net/projects/mpv-player-windows/files/latest/download
'''
zurl = [
    'https://sourceforge.com/projects/mpv-player-windows/files/64bit/mpv-x86_64-20260419-git-06f4ce7.7z/download',
    'https://sourceforge.net/projects/mplayerwin/files/MPlayer-MEncoder/r38151/mplayer-svn-38151-x86_64.7z/download'
]

def win_press_any_key_to_uninstall():
    print('\nTo complete the process you will have to [red]execute a batch file[/red].')
    print('Windows Explorer will open the location of the batch file to run.\n')
    print('Please double click\n')
    print('    [bold green]uninstall.bat[/bold green]\n')
    print('to remove [magenta]PyRadio[/magenta] from your system.\n')
    print('After you are done, you can delete the folder it resides in.\n')
    print('Press any key to exit...', end='', flush=True)

    getwch()

    with as_file(files("pyradio").joinpath("install.py")) as tmp_path:
        subprocess.call(
            [sys.executable, str(tmp_path), "-R"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

def win_print_exe_paths():
    from .install import fix_pyradio_win_exe
    exe = fix_pyradio_win_exe()
    if exe[0] and exe[1]:
        print('[magenta]PyRadio[/magenta] EXE files:')
        print(f'  System:\n    [red]{exe[0]}[/red]')
        print(f'  User:\n    [green]{exe[1]}[/green]')
    else:
        print('[magenta]PyRadio[/magenta] EXE file:')
        if exe[0]:
            print(f'  [green]{exe[0]}[/green]')
        else:
            print(f'  [green]{exe[1]}[/green]')
    # doing it this way so that python2 does not break (#153)
    print('\nPress any key to exit...', end='', flush=True)
    getwch()

def press_any_key_to_continue():
    print('\nPress any key to exit...', end='', flush=True)
    getwch()

def install_module(a_module, do_not_exit=False, print_msg=True):
    if print_msg:
        print('Installing module: [green]' + a_module + '[/green]')
    for count in range(1,6):
        ret = subprocess.call('python -m pip install --upgrade ' + a_module,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)
        if ret == 0:
            break
        if count < 5:
            if print_msg:
                print(f'  Download failed. Retrying [magenta]{count + 1}[/magenta]/[red]5[/red]')
        else:
            if print_msg:
                print('Failed to download module...\nPlease check your internet connection and try again...')
            else:
                print('Failed to download module "[magenta]{}[/magenta]"...\nPlease check your internet connection and try again...').format(a_module)
            if do_not_exit:
                return False
            sys.exit(1)
        return True

try:
    from rich import print
except:
    install_module('rich')
    from rich import print

def find_pyradio_win_exe():
    ''' find pyradio EXE files

        Return (system_exe, user_exe)
    '''
    exe = [None, None]
    for a_path in site.getsitepackages():
        an_exe = join(a_path, 'Scripts' , 'pyradio.exe')
        if exists(an_exe):
            exe[0] = an_exe
            break
    an_exe = join(site.getuserbase(), 'Scripts' , 'pyradio.exe')
    if exists(an_exe):
        exe[1] = an_exe
    # print(exe)
    return exe

def _is_player_in_path(a_player):
    ''' Return player's path in PATH variable
        If player not in PATH, return None
        Makes sure the path is local to user
        and player EXE exists

        Parameter:
            a_player: 1=mpv, 2=mplayer
    '''
    a_player -= 1
    in_path = None
    pl = ('mpv', 'mplayer')
    for a_path in environ['PATH'].split(';'):
        if a_path.endswith(pl[a_player]):
            in_path = a_path
            break
    #print('in_payh: {}'.format(in_path))
    if in_path:
        if environ['USERPROFILE'] not in a_path:
            return None
        if not exists(join(in_path, pl[a_player] + '.exe')):
            return None
    return in_path

def _get_output_folder(package, output_folder=None, do_not_exit=False):
    if output_folder is None:
        a_path = _is_player_in_path(package)
        if a_path:
            sp = a_path.split(sep)
            output_folder = sep.join(sp[:-1])
        else:
            output_folder = join(environ['APPDATA'], 'pyradio')
        # rename mpv if already there
        if not exists(output_folder):
            # create dir
            makedirs(output_folder, exist_ok=True)
            if not exists(output_folder):
                print(f'Failed to create folder: "[magenta]{output_folder}[/magenta]"')
                if do_not_exit:
                    return None
                sys.exit(1)
    return output_folder

def _get_out_file(output_folder, package=1):
    count = 0
    p_name=('mpv-latest', 'mplayer-latest')
    out_file = join(output_folder, f'{p_name[package]}.7z')
    while True:
        if exists(out_file):
            count += 1
            out_file = join(output_folder, f'{p_name[package]}-{count}.7z')
        else:
            break
    return join(output_folder, out_file)

def download_seven_zip(output_folder):
    PR = (
        join(getenv('PROGRAMFILES'), '7-Zip', '7z.exe'),
        join(getenv('PROGRAMFILES') + ' (x86)', '7-Zip', '7z.exe')
    )
    if exists(PR[0]) or exists(PR[1]):
        return

    url = 'https://sourceforge.net/projects/sevenzip/files/latest/download'

    out_file = join(output_folder, '7-Zip_latest.exe')

    print('[magenta]7-Zip not found...\n[green]Downloading...[/green]')
    try:
        urlretrieve(url, filename=out_file)
    except:
        print('[red]Failed to download 7-Zip...[/red]')
        print('Please check your internet connection and try again...')
        print('\nIn case you want to [green]install 7-Zip manually[/green],')
        print('go to [magenta]https://www.7-zip.org/[/magenta] to get it...')
        sys.exit(1)

    print('\n[bold]PyRadio installation will resume as soon as\nyou complete the installation of 7-Zip...[/bold]')

    subprocess.call(
        out_file,
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )

def get_latest_x86_64_mplayer_url():
    url = purl[1]
    try:
        r = requests.get(url)
    except:
        return None
    if r.status_code == 200:
        # print(r.text)
        sp = r.text.split('<tr title="')
        for n in sp:
            if n.startswith('r'):
                rev = n.split('"')[0][1:]
                try:
                    int(rev)
                except ValueError:
                    return None
                existing = zurl[1].split('-svn-')[1].split('-')[0]
                # print('     rev = ' + rev[1:])
                # print('existing = ' + existing[1:])
                if existing == rev:
                    rev = None
                else:
                    zurl[1] = zurl[1].replace(existing, rev[1:])
                r.close()
                return rev
    r.close()
    return None

def get_latest_x86_64_mpv_url():
    url = purl[0]
    try:
        r = requests.get(url)
    except:
        return None
    if r.status_code == 200:
        # print(r.text)
        sp = r.text.split('"url":"')
        for n in sp:
            if n.startswith('/project') and '-v3' not in n:
                # print('item: {}'.format(i))
                # print(n)
                # print('\n\n\n')
                r.close()
                return 'https://sourceforge.com' + n.split('"')[0] + 'download'
    r.close()
    return None

def download_player(output_folder=None, package=1, do_not_exit=False):
    # Parameters
    #   output_folder   : where to save files
    #   package         : 0: mpv, 1: mplayer
    package -= 1
    if package == 0:
        print('Downloading [magenta]MPV[/magenta] ([green]latest[/green])...')
    else:
        print('Downloading [magenta]MPlayer[/magenta] ([green]latest[/green])...')

    mpv_url = None
    if package == 0:
        if is_cpu_mpv_v3_compatible():
            zurl[0] = 'https://sourceforge.net/projects/mpv-player-windows/files/latest/download'
        else:
            mpv_url = get_latest_x86_64_mpv_url()
            if mpv_url:
                zurl[0] = mpv_url
    elif package == 1:
        get_latest_x86_64_mplayer_url()
    output_folder = _get_output_folder(
        output_folder=output_folder,
        package=package,
        do_not_exit=do_not_exit)
    if output_folder is None:
        return False
    if True == False and package == 0 and \
            exists(join(output_folder, 'mpv', 'updater.bat')):
        chdir(join(output_folder, 'mpv'))
        startfile('updater.bat')
    else:
        print(f'    from  "[plum4]{purl[package]}[plum4]"')
        print(f'    into  "[magenta]{output_folder}[/magenta]"')

        out_file = _get_out_file(output_folder, package)
        session = requests.Session()
        for count in range(1,6):
            try:
                r = session.get(zurl[package])
                r.raise_for_status()
                break
            except requests.exceptions.RequestException:
                if count < 5:
                    print(f'  Download failed. Retrying [magenta]{count + 1}[/magenta]/[red]5[/red]')
                else:
                    print('[red]Failed to download player...[/red]\nPlease check your internet connection and try again...')
                    if do_not_exit:
                        return False
                    sys.exit(1)
        print(f'  Saving: "{out_file}"')
        try:
            with open(out_file, 'wb') as f:
                f.write(r.content)
        except:
            print('[red]Failed to write archive...[/red]\nPlease try again later...')
            if do_not_exit:
                return False
            sys.exit(1)

        print('Extracting archive...')
        if package == 0:
            download_seven_zip(output_folder)

        if not HAVE_PYUNPACK:
            install_module('pyunpack', print_msg=False)
        from pyunpack import Archive

        count = 0
        while True:
            patool_exec = join(site.USER_SITE.replace('site-packages', 'Scripts'), 'patool')
            if exists(patool_exec):
                break
            patool_exec = glob.glob(join(dirname(environ['APPDATA']), '**', 'patool.exe'), recursive=True)
            if patool_exec:
                patool_exec = patool_exec[0]
                break
            install_module('patool', print_msg=False)
            count += 1
            if count > 2:
                break
        try:
            Archive(out_file).extractall(join(output_folder, 'mpv' if package==0 else ''),
                auto_create_dir=True,
                patool_path=patool_exec)
        except:
            file_only = basename(out_file)
            player_name = 'mpv' if package == 0 else 'mplayer'
            print(f'''Failed to extract the archive...

    You will have to install the player [red]MANUALLY[/red]!!!

    PyRadio's configuration folder will open now,
    along with the archive named "{basename(out_file)}".''')
            if player_name == 'mpv':
                if exists(join(output_folder, 'mpv')):
                    print('''    Please extract the archive in the "[red]mpv[/red]" folder
    (overwriting any existing files).''')
                else:
                    print('''    Please create a folder named "[red]mpv[/red]" and extract
    the archive there.''')
            else:
                # mplayer
                if exists(join(output_folder, 'mplayer')):
                    print('''    Please delete the "[red]mplayer[/red]" folder, extract
    the archive and rename the resulting folder
    to "[red]mplayer[/red]".''')
                else:
                    print('''
    Please extract the archive and rename the resulting
    folder to "[red]mplayer[/red]".''')

            print('Press any key to continue...')

            getwch()
            if player_name == 'mpv':
                startfile(join(dirname(out_file), 'mpv'))
            else:
                startfile(dirname(out_file))
            startfile(out_file)

            '''
            if player_name == 'mpv':
                while not exists(join(output_folder, 'mpv', 'updater.bat')):
                    sleep(1)
                chdir(join(output_folder, 'mpv'))
                startfile('updater.bat')
            '''

            if do_not_exit:
                return False
            sys.exit(1)

    if not _post_download(package, output_folder, do_not_exit):
        return False
    try:
        remove(out_file)
    except:
        pass
    return True

def is_cpu_mpv_v3_compatible(print_flags=False):
    try:
        import cpuinfo
    except:
        install_module('py-cpuinfo')
    from cpuinfo import get_cpu_info
    cpu = get_cpu_info()

    ''' Source: Difference between x86_64 and x86_64 v3
        https://github.com/shinchiro/mpv-winbuild-cmake/wiki/Difference-between-x86_64-and-x86_64-v3
    '''
    mpv_v3_cpu_flags = (
    'avx',
    'avx2',
    'bmi',
    'bmi2',
    'fma',
    'lzcnt',
    'movbe',
    'sse3',
    'sse4',
    'sse4_1',
    'sse4_2',
    'ssse3',
    'xsave',
    'sahf',
    'mwait',
    'crc32',
    'cx16',
    'popcnt',
    'f16c'
    )

    count = 0

    for n in mpv_v3_cpu_flags:
        if n in cpu['flags']:
            count += 1
            if print_flags:
                print(n, '\t', 'ok')
        else:
            if print_flags:
                print(n, '\t', 'nok')

    return len(mpv_v3_cpu_flags) == count

def _post_download(package, output_folder, do_not_exit):

    # rename MPlayer directory
    if package == 1:
        sleep(5)
        mplayer_dir_found = False
        extracted_dirname = None
        dir_list = listdir(output_folder)
        for a_file in dir_list:
            if a_file == 'mplayer':
                mplayer_dir_found = True
            elif a_file.lower().startswith('mplayer-svn') and \
                    isdir(join(output_folder, a_file)):
                extracted_dirname = a_file

        # rename extracted dir to mplayer
        if extracted_dirname:
            extracted_dirname = join(output_folder, extracted_dirname)
            mplayer_final_dir = join(output_folder, 'mplayer')
            mplayer_old_dir = join(output_folder, 'mplayer.old')

            if mplayer_dir_found:
                if exists(mplayer_old_dir):
                    try:
                        rmtree(mplayer_old_dir)
                    except OSError:
                        print(f'Failed to remove "[green]{mplayer_old_dir}[/green]"\nPlease close all programs and try again...')
                        if do_not_exit:
                            return False
                        sys.exit(1)
                try:
                    replace(mplayer_final_dir, mplayer_old_dir)
                except:
                    print(f'Failed to rename folder "[green]{mplayer_final_dir}[/green]"\n      to "[magenta]{mplayer_old_dir}[/magenta]"...\nPlease close all open programs and try again...')
                    if do_not_exit:
                        return False
                    sys.exit(1)
            try:
                replace(join(output_folder, extracted_dirname), join(output_folder, 'mplayer'))
            except:
                print(f'Failed to rename folder "[green]{extracted_dirname}[/green]" to\n      "[magenta]{mplayer_final_dir}[/magenta]"...\nPlease close all open programs and try again...')
                if do_not_exit:
                    return False
                sys.exit(1)

        else:
            print('[red]Extracted folder not found...[/red]\nPlease try again later...')
            if do_not_exit:
                return False
            sys.exit(1)
    return True

def install_player(output_folder=None, package=0, do_not_exit=False):
    while True:
        in_path = [None, None, None]
        to_do = ['[bold red]1[/bold red]. Install', '[bold red]2[/bold red]. Install', '[green]VLC[/green] media player is not installed']
        from_path = ['', '']
        for n in range(0, 2):
            in_path[n] = _is_player_in_path(n)
            if in_path[n]:
                to_do[n] = f'[bold red]{n + 1}[/bold red]. Update'
                from_path[n] = ' (found in [magenat]PATH[/magenta])'
        if in_path[0] is None:
            in_path[0] = find_mpv_on_windows()
        if in_path[1] is None:
            in_path[1] = find_mplayer_on_windows()

        if in_path[0] == 'mpv':
            in_path[0] = None
        if in_path[1] == 'mplayer':
            in_path[1] = None

        for n in range(0, 2):
            if in_path[n]:
                to_do[n] = f'[bold red]{n + 1}[/bold red]. Update'
        if find_vlc_on_windows():
            to_do[2] = '[green]VLC[/green] media player is already installed.\n[bold red]      It is not recommended to be used!!![/bold red]'
        #print(in_path)
        #print(to_do)
        #print(from_path)

        #print('\nDo you want to download a media player now? (Y/n): ', end='', flush=True)
        #x = getwch()
        #print(x)
        x = 'y'
        if in_path[0]:
            best_choise = ''
        else:
            best_choise = '([yellow]best choise[/yellow])'
        if x in ('y', '\n', '\r'):
            x = ''
            msg = '''
Please select an action:
    {0} [green]MPV[/green]{1}      {2}
    {3} [green]MPlayer[/green]{4}'''


            print(msg.format(to_do[0], from_path[0],
                best_choise, to_do[1], from_path[1]
            ))
            msg ='''
    [plum4]Note:[/plum4]
      {}
    '''
            opts = []
            prompt = ''
            all_uninstall = False
            if in_path[0] is None and in_path[1] is None:
                opts = ['0', '1', '2', 'q']
                prompt = 'Press [bold red]1[/bold red], [bold red]2[/bold red] or [bold red]q[/bold red] to Cancel: '
            elif in_path[0] is not None and in_path[1] is not None:
                print('\n    [bold red]3[/bold red]. Uninstall [green]MPV[/green]')
                print('    [bold red]4[/bold red]. Uninstall [green]MPlayer[/green]')
                opts = ['0', '1', '2', '3', '4', 'q']
                prompt = 'Press [bold red]1[/bold red], [bold red]2[/bold red], [bold red]3,[/bold red] [bold red]4[/bold red] or [bold red]q[/bold red] to Cancel: '
            else:
                if in_path[0] is not None:
                    print('\n    [bold red]3[/bold red]. Uninstall [green]MPV[/green]')
                else:
                    print('\n    [bold red]3[/bold red]. Uninstall [green]MPlayer[/green]')
                opts = ['0', '1', '2', '3', 'q']
                prompt = 'Press [bold red]1[/bold red], [bold red]2[/bold red], [bold red]3[/bold red] or [bold red]q[/bold red] to Cancel: '
                all_uninstall = True

            print(msg.format(to_do[2]))

            while x not in opts:
                print(prompt,  end='', flush=True)
                x = getwch()
                print(x)

            # ok, parse response
            if x in ('0', 'q'):
                clean_up()
                return
            if x in ('1', '2'):
                # install ot update
                download_player(package=int(x))
                print('\n\n')
            elif x == '3':
                # find out which player to wuninstall
                print('uninstall [green]mplayer[/green] or [green]mpv[/green]')
                print('\n\n')
            elif x == '4':
                # uninstall mplayer
                print('uninstall [green]mplayer[/green]')
                print('\n\n')

def install_pylnk(a_path, do_not_exit=False):
    print('    Downloading [green]pylnk[/green]...')
    session = requests.Session()
    for count in range(1,6):
        try:
            r = session.get('https://github.com/strayge/pylnk/archive/refs/heads/master.zip')
            r.raise_for_status()
            break
        except requests.exceptions.RequestException:
            if count < 5:
                print(f'      Download failed. Retrying [magenta]{count + 1}[/magenta]/[red]5[/red]')
            else:
                print('    Failed to download [green]pylnk[/green]...\nPlease check your internet connection and try again...')
                if do_not_exit:
                    return False
                sys.exit(1)
    try:
        with open(join(a_path, 'pylnk.zip'), 'wb') as f:
            f.write(r.content)
    except:
        print('    Failed to write archive...\nPlease try again later...')
        if do_not_exit:
            return False
        sys.exit(1)

    print('    Installing [green]pylnk...[/green]')
    subprocess.call('python -m pip install ' + join(a_path, 'pylnk.zip'),
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL)

    remove(join(a_path, 'pylnk.zip'))

def clean_up(print_msg=True):
    if print_msg:
        print('Cleaning up makedepend modules...')
    for n in ('pyunpack', 'patool', 'pylnk3', 'EasyProcess'):
        subprocess.call('python -m pip uninstall -y ' + n,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL)

def get_path(exe):
    out_exe = ''
    chk = []

    for n in site.getsitepackages():
        # print('adding: "{}"'.format(join(n, exe)))
        # print('adding: "{}"'.format(join(n, 'Scripts', exe)))
        chk.append(join(n, exe))
        chk.append(join(n, 'Scripts', exe))
    # print('------------------------')
    x = site.getusersitepackages()
    if isinstance(x, str):
        # print('adding: "{}"'.format(join(x, exe)))
        # print('adding: "{}"'.format(join(x, 'Scripts', exe)))
        chk.append(join(x, exe))
        chk.append(join(x, 'Scripts', exe))
        # print('adding: "{}"'.format(join(x, exe)).replace('\site-packages', ''))
        # print('adding: "{}"'.format(join(x, 'Scripts', exe)).replace('\site-packages', ''))
        chk.append(join(x, exe).replace(r'\site-packages', ''))
        chk.append(join(x, 'Scripts', exe).replace(r'\site-packages', ''))
    else:
        for n in site.getusersitepackages():
            # print('adding: "{}"'.format(join(n, exe)))
            # print('adding: "{}"'.format(join(n, 'Scripts', exe)))
            chk.append(join(n, exe))
            chk.append(join(n, 'Scripts', exe))
    # print('------------------------')
    for n in site.PREFIXES:
        # print('adding: "{}"'.format(join(n, exe)))
        # print('adding: "{}"'.format(join(n, 'Scripts', exe)))
        chk.append(join(n, exe))
        chk.append(join(n, 'Scripts', exe))
    # for n in range(0,4):
    #     print('')
    # for n in chk:
    #     print(n)
    # print('------------------------')
    for n in chk:
        # print('checking: "{}'.format(n))
        if exists(n):
            return n
    return ''

def get_pyradio():
    return get_path('pyradio.exe')

def get_pylnk():
    return get_path('pylnk3.exe')

def create_pyradio_link():
    pyradio_exe = 'pyradio'
    pyradio_exe = get_pyradio()
    pylnk_exe = get_pylnk()
    # print('pyradio_exe = "{}"'.format(pyradio_exe))
    # print('pylnk_exe = "{}"'.format(pylnk_exe))
    icon = join(environ['APPDATA'], 'pyradio', 'help', 'pyradio.ico')
    # print('icon = "{}"'.format(icon))
    link_path = join(environ['APPDATA'], 'pyradio', 'help', 'PyRadio.lnk')
    # print('link_path = "{}"'.format(link_path))
    workdir = join(environ['APPDATA'], 'pyradio')
    # print('workdir = "{}"'.format(workdir))
    # print('*** Updating Dekstop Shortcut')
    if not exists(workdir):
        makedirs(workdir, exist_ok=True)
        if not exists(workdir):
            print('Cannot create "' + workdir + '"')
            sys.exit(1)
    if not exists(pylnk_exe):
        install_pylnk(workdir)
    pylnk_exe = get_pylnk()
    # print('pylnk_exe = "{}"'.format(pylnk_exe))
    # cmd = pylnk_exe + ' c --icon ' + icon + ' --workdir ' + workdir \
    #     + ' ' + pyradio_exe + ' ' + link_path
    # print('cmd = "{}"'.format(cmd))
    subprocess.Popen(
        [pylnk_exe, 'c', '--icon', icon, '--workdir', workdir, pyradio_exe, link_path],
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL
    )

def install_pyradio_link():
    from shutil import copy
    desktop = getenv('DESKTOP')
    user_profile = getenv('USERPROFILE')
    appdata = getenv('APPDATA')
    to_desktop = desktop if desktop is not None else join(user_profile, 'desktop')
    to_start_menu = join(appdata, 'Microsoft', 'Windows', 'Start Menu', 'Programs')

    if exists(to_desktop):
        copy(join(appdata, 'pyradio', 'help', 'PyRadio.lnk'),
             join(to_desktop, 'PyRadio.lnk')
             )

    if exists(to_start_menu):
        copy(join(appdata, 'pyradio', 'help', 'PyRadio.lnk'),
             join(to_start_menu, 'PyRadio.lnk')
             )

def find_and_remove_recording_data(data_dir):
    threading.Thread(
        target=find_and_remove_recording_data_thread(data_dir)
        ).start()

def find_and_remove_recording_data_thread(data_dir):
    files = glob.glob(join(data_dir, 'tmp_*.mkv'))
    txt = glob.glob(join(data_dir, '*.txt'))
    xml = glob.glob(join(data_dir, '*.xml'))

    for i in range(len(txt)-1, -1, -1):
        if txt[i].startswith('search-'):
            txt.pop(i)
    for n in txt, xml:
        if n:
            files.extend(n)
    if files:
        for n in files:
            try:
                remove(path.join(data_dir, n))
            except:
                pass

if __name__ == '__main__':
    print('\n\n[red]----[green]====  [magenta]MPV Media Player Installation  [green]====[red]----[/red]')
    download_player(package=1)
    print('[red]----[green]====  [magenta]MPV Media Player Installed  [green]====[red]----[/red]')
    # _post_download(1, "C:\\Users\\spiros\\AppData\\Roaming\\pyradio")
    # download_player(package=0)
    #install_player()

    # install_pylnk("C:\\Users\\spiros")
    #create_pyradio_link()
    # find_pyradio_win_exe()



# -*- coding: utf-8 -*-
import sys
import hashlib
import csv

import locale
locale.setlocale(locale.LC_ALL, "")

class CompareFiles():
    ''' Class to compare two files  by content '''

    def __init__(self, fname1, fname2):
        ''' Just provide two files' pathname and use equal of different
            properties to get the result.
            They return True or False, or None if an error opening any
            one of the files occurs.
        '''

        self.fname1 = fname1
        self.fname2 = fname2
        self.sha1 = None
        self.sha2 = None

    @property
    def equal(self):
        ''' Checks if files are equal
            Returns
                True    files are equal
                False   files are different
                None    error reading files
        '''
        if self._compare_files():
            return self.sha1 == self.sha2
        return None

    @equal.setter
    def equal(self, val):
        raise ValueError('This property is read-only')

    @property
    def different(self):
        ''' Checks if files are different
            Returns
                True    files are different
                False   files are equal
                None    error reading files
        '''
        if self._compare_files():
            return not self.sha1 == self.sha2
        return None

    @different.setter
    def different(self, val):
        raise ValueError('This property is read-only')

    def _compare_files(self):
        try:
            self.sha1 = self._sha512(self.fname1)
            self.sha2 = self._sha512(self.fname2)
            return True
        except:
            self.sha1 = None
            self.sha2 = None
            return False

    def _sha512(self, fname):
        hash_sha512 = hashlib.sha512()
        with open(fname, "rb") as f:
            for chunk in iter(lambda: f.read(4096), b""):
                hash_sha512.update(chunk)
                return hash_sha512.hexdigest()

    def read_file(self, a_file):
        stations_read = []
        with open(a_file, 'r', encoding='utf-8') as cfgfile:
            try:
                for row in csv.reader(filter(lambda row: row[0]!='#', cfgfile), skipinitialspace=True):
                    if not row:
                        continue
                    try:
                        name, url = [s.strip() for s in row]
                        stations_read.append([name, url, '', ''])
                    except:
                        try:
                            name, url, enc = [s.strip() for s in row]
                            stations_read.append([name, url, enc, ''])
                        except:
                            name, url, enc, onl = [s.strip() for s in row]
                            stations_read.append([name, url, enc, onl])
            except:
                stations_read = []
                return None
        return stations_read

if __name__ == '__main__':
    files = (
        '/home/spiros/.config/pyradio/stations.csv',
        '/home/spiros/projects/my-gits/pyradio/pyradio/stations.csv'
    )
    cmp = CompareFiles(files[0], files[1])

    res = cmp.different
    if res is None:
        print('Error opening files')
    else:
        if res:
            print('Files are different')
        else:
            print('Files are equal')
            sys.exit()

    first_list = cmp.read_file(files[0])
    second_list = cmp.read_file(files[1])

    # diff = [x for x in first_list if x not in second_list] + [x for x in second_list if x not in first_list]
    old = [x for x in first_list if x not in second_list]
    new = [x for x in second_list if x not in first_list]


    for n in old:
        print(n)

    print('\n\n\n')

    for n in new:
        print(n)



# -*- coding: utf-8 -*-
import locale
import subprocess
import logging
from sys import platform
from os import path, environ, listdir, getenv
from shutil import which
from platform import system
from .install import get_a_linux_resource_opener
try:
    import psutil
    HAVE_PSUTIL = True
except:
    HAVE_PSUTIL = False

logger = logging.getLogger(__name__)

locale.setlocale(locale.LC_ALL, "")

'''
    0 : perform detection
    1 : Graphical Environment running
    2 : Graphical Environment not running
'''
HAS_GRAPHICAL_ENV = 0

def convert_to_md(a_file):
    tmp_file = a_file[:-4] + 'md'
    return tmp_file if path.exists(tmp_file) else a_file

def is_graphical_environment_running():
    global HAS_GRAPHICAL_ENV
    this_system = system()
    if this_system in ('Darwin', 'Windows'):
        HAS_GRAPHICAL_ENV = 1
        return True
    if getenv('TERM') == 'linux':
        HAS_GRAPHICAL_ENV = 2
        return False

    if HAS_GRAPHICAL_ENV == 1:
        return True
    if HAS_GRAPHICAL_ENV == 2:
        return False

    if which('pgrep'):
        # Check if Xorg is running
        xorg_process = subprocess.run(['pgrep', '-x', 'Xorg'], stdout=subprocess.PIPE)
        if xorg_process.returncode == 0:
            HAS_GRAPHICAL_ENV = 1
            return True
        # Check if Wayland is running
        wayland_process = subprocess.run(['pgrep', '-x', 'wayland'], stdout=subprocess.PIPE)
        if wayland_process.returncode == 0:
            HAS_GRAPHICAL_ENV = 1
            return True
    elif path.exists('/proc'):
        for pid in listdir('/proc'):
            if pid.isdigit():
                try:
                    with open(f'/proc/{pid}/cmdline', 'rb') as f:
                        cmdline = f.read().decode().split('\x00')
                        if 'Xorg' in ' '.join(cmdline) or \
                            'wayland' in ' '.join(cmdline):
                            return True
                except FileNotFoundError:
                    continue
    elif HAVE_PSUTIL:
        for proc in psutil.process_iter(['pid', 'name']):
            if proc.info['name'] == 'Xorg' \
                    or proc.info['name'] == 'wayland':
                HAS_GRAPHICAL_ENV = 1
                return True

    # Check if DISPLAY environment variable is set
    if 'DISPLAY' in environ:
        HAS_GRAPHICAL_ENV = 1
        return True

    HAS_GRAPHICAL_ENV = 2
    return False

class HtmlHelp():

    def __init__(self):
        self._files = ('index.html', 'radio-browser.html')

        if platform.lower().startswith('win'):
            self._paths = (path.join(path.expandvars('%APPDATA%'), 'pyradio', 'help'), )
        else:
            self._paths = (
                '/usr/share/doc/pyradio',
                '/usr/local/share/doc/pyradio',
                path.join(path.expanduser('~'), '.local/share/doc/pyradio')
            )

        for a_path in self._paths:
            if path.exists(a_path):
                self._path = a_path
                break

    def open_file(self, linux_resource_opener=None, browser=False):
        a_file = self._files[1] if browser else self._files[0]
        self._open_file(a_file, linux_resource_opener=linux_resource_opener)

    def open_filename(self,a_file, linux_resource_opener=None):
        self._open_file(a_file, linux_resource_opener=linux_resource_opener)

    def _open_file(self, a_file, linux_resource_opener=None):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('HtmlHelp: opening "{}"'.format(path.join(self._path, a_file)))
        this_platform = platform.lower()
        if this_platform.startswith('win'):
            from os import startfile
            startfile(path.join(self._path, a_file))
        else:
            if this_platform.startswith('darwin'):
                cmd = [which('open'),  path.join(self._path, a_file)]
            else:
                ''' linux '''
                # if is_graphical_environment_running():
                #     return
                if linux_resource_opener is None:
                    tool = get_a_linux_resource_opener()
                else:
                    tool = linux_resource_opener
                if tool is None:
                    if logger.isEnabledFor(logging.INFO):
                        logger.info('HtmlHelp: Cannot find a run tool for Linux!')
                    return
                if isinstance(tool, str):
                    tool = tool.split(' ')
                # a_file = convert_to_md(a_file)
                cmd = [*tool , path.join(self._path, a_file)]
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug('HtmlHelp: executing: "{}"'.format(cmd))
            try:
                subprocess.Popen(
                       cmd,
                       shell=False,
                       stdout=subprocess.PIPE,
                       stderr=subprocess.PIPE, close_fds=True)
            except (FileNotFoundError, PermissionError):
                pass




#!/usr/bin/env python3
"""
TTSManager - Cross-platform Text-to-Speech with Smart Queue Management
Advanced TTS manager with priority-based queue, title preservation,
volume debouncing and cross-platform support.
"""

import platform
import subprocess
import threading
import time
import logging
import os
import shlex
import queue
from enum import Enum
from datetime import datetime
if platform.system().lower().startswith('win'):
    try:
        import win32com.client
    except:
        pass
from .common import M_STRINGS
from .tts_text import tts_transform_to_string

logger = logging.getLogger(__name__)

class Priority(Enum):
    """Priority levels for speech requests"""
    NORMAL = 1          # Interruptible - navigation, titles
    HIGH = 2            # Non-interruptible - critical alerts, playback status
    NAVIGATION = 3      # HIGH-like but self-interruptible
    HELP = 4            # Help messages
    DIALOG = 5          # Interruptible HIGH - dialog messages

class Context(Enum):
    """Context setting for speech requests"""
    LIMITED = 1
    WINDOW = 2
    ALL = 3

class TTSState(Enum):
    """TTS system states"""
    IDLE = 0
    SPEAKING = 1
    SHUTTING_DOWN = 2

class TTSConfig:
    """Configuration manager for TTS commands"""

    def __init__(self):
        self.config_file = self._get_config_path()
        self.config = self._load_config()

    def _get_config_path(self):
        """Get config file path for current platform"""
        if platform.system() == "Windows":
            return os.path.expanduser("~\\AppData\\Local\\pyradio\\tts.conf")
        return os.path.expanduser("~/.config/pyradio/tts.conf")

    def _load_config(self):
        """Load configuration from file with cross-platform support"""
        # Default configurations
        defaults = {
            'COMMAND_LINUX': 'spd-say',
            'COMMAND_WINDOWS': 'powershell -Command "Add-Type -AssemblyName System.Speech; $s=New-Object System.Speech.Synthesis.SpeechSynthesizer; $s.Speak(\\\"{}\\\")"',
            'COMMAND_MACOS': 'say',
            'RESTART_LINUX': 'systemctl --user restart speech-dispatcher',
            'TITLE_TOKEN': M_STRINGS['title_'],
            'RESET_TOKENS': 'Playing,Initializing,Connecting,Error:'
        }

        try:
            with open(self.config_file, 'r', encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    # Skip comments and empty lines
                    if not line or line.startswith('#'):
                        continue
                    # Parse key=value
                    if '=' in line:
                        key, value = line.split('=', 1)
                        key = key.strip()
                        value = value.strip()
                        if key in defaults:
                            defaults[key] = value
            if logger.isEnabledFor(logging.INFO):
                logger.info(f"Loaded TTS config from {self.config_file}")
        except FileNotFoundError:
            if logger.isEnabledFor(logging.INFO):
                logger.info("No TTS config file found, using defaults")
        except Exception as e:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning(f"Error reading TTS config: {e}")

        return defaults

    def get_command(self, system, text):
        """Get the appropriate command for the current system and text"""
        if system == "Linux":
            cmd_template = self.config['COMMAND_LINUX']
        elif system == "Windows":
            cmd_template = self.config['COMMAND_WINDOWS']
        elif system == "Darwin":
            cmd_template = self.config['COMMAND_MACOS']
        else:
            return None

        # Replace placeholder or append text
        if '{}' in cmd_template:
            return cmd_template.replace('{}', text)
        # Use shlex.split to properly handle quoted arguments
        base_cmd = shlex.split(cmd_template)
        base_cmd.append(text)
        return base_cmd

    def get_linux_restart_commands(self):
        """Get Linux restart commands as a list of command lists"""
        restart_str = self.config['RESTART_LINUX']
        commands = []

        if '&&' in restart_str:
            # Multiple commands separated with a &&
            for cmd in restart_str.split('&&'):
                cmd = cmd.strip()
                if cmd:
                    commands.append(shlex.split(cmd))
        else:
            # Single command
            commands.append(shlex.split(restart_str))

        return commands

    def get_title_token(self):
        """Get the title token for identifying title messages"""
        return self.config['TITLE_TOKEN']

    def get_reset_tokens(self):
        """Get the reset tokens that trigger title reset"""
        return [token.strip() for token in self.config['RESET_TOKENS'].split(',')]

class TTSRequest:
    """Represents a TTS speech request"""

    def __init__(self, text, priority=Priority.NORMAL):
        self.text = text
        self.priority = priority
        self.timestamp = time.time()

class TTSBase:
    """Base class for TTS implementations"""

    def __init__(self, config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start):
        self.config = config
        self.volume = volume
        self.rate = rate
        self.pitch = pitch
        self.verbosity = verbosity
        self.speak_volume = speak_volume
        self.speak_volume_start = speak_volume_start
        self.system = platform.system()
        self.state = TTSState.IDLE
        self._current_process = None
        self._lock = threading.RLock()
        self.speech_delay = 0.3  # Anti-stutter delay

        # External stop mechanism for DIALOG and NAVIGATION interruption
        self._external_stop_requested = threading.Event()
        self._external_stop_lock = threading.RLock()
        self._navigation_stop_requested = threading.Event()  # NEW: Specific for NAVIGATION

    def request_external_stop(self, priority=Priority.DIALOG):
        """Request external stop - to be checked during DIALOG or NAVIGATION execution"""
        logger.error('priority  = {}'.format(priority.name))
        with self._external_stop_lock:
            if priority == Priority.NAVIGATION:
                logger.error('setting navigation stop request')
                self._navigation_stop_requested.set()
            else:
                logger.error('setting external stop request')
                self._external_stop_requested.set()

    def _clear_external_stop(self, priority=None):
        """Clear external stop flag for specific priority"""
        with self._external_stop_lock:
            if priority == Priority.NAVIGATION:
                self._navigation_stop_requested.clear()
            elif priority in (Priority.DIALOG, Priority.HIGH):
                self._external_stop_requested.clear()
            else:
                self._external_stop_requested.clear()
                self._navigation_stop_requested.clear()

    def _should_stop_externally(self, priority):
        """Check if external stop was requested"""
        result = False
        if priority in (Priority.DIALOG, Priority.HIGH):
            with self._external_stop_lock:
                result = self._external_stop_requested.is_set()
                if result:
                    logger.error(f"macOS DEBUG: _should_stop_externally returning TRUE for {priority.name}")
        elif priority == Priority.NAVIGATION:
            with self._external_stop_lock:
                result = self._navigation_stop_requested.is_set()

        return result

class TTSLinux(TTSBase):
    """Linux TTS implementation using user-configured command"""

    def __init__(self, config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start):
        super().__init__(config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start)
        self.retry_count = 0
        self.max_retries = 2

    def _calculate_volume(self):
        volume = int(self.volume())
        return str(2 * volume - 100)

    def _execute_speech(self, text, priority=Priority.NORMAL):
        """Execute speech with priority-based blocking behavior"""
        try:
            # Clear external stop flag for the given priority
            self._clear_external_stop(priority)

            # Calculate volume for spd-say (-100 to 100)
            volume = self._calculate_volume()

            # Determine command based on priority
            if priority in (Priority.HIGH, Priority.DIALOG, Priority.NAVIGATION):
                # HIGH/DIALOG/NAVIGATION priority: blocking execution with -w flag
                cmd = ['spd-say', '-l', 'en', '-i', volume,
                       '-r', self.rate(), '-p', self.pitch(), '-w', text]
                logger.error(f'Linux: Blocking speech with -w flag for {priority.name}')
            else:
                # NORMAL priority: non-blocking execution
                cmd = ['spd-say', '-l', 'en', '-i', volume,
                       '-r', self.rate(), '-p', self.pitch(), text]
                logger.error(f'Linux: Non-blocking speech for {priority.name}')

            # Execute the command
            logger.error(f'===> executing: "{cmd}" with {priority.name = }')

            # For blocking priorities (HIGH/DIALOG/NAVIGATION), we need to handle interruption
            if priority in (Priority.HIGH, Priority.DIALOG, Priority.NAVIGATION):
                # Start the process
                process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                    text=True
                )

                # Monitor the process for completion or interruption
                while process.poll() is None:
                    # Small sleep to avoid busy waiting
                    time.sleep(0.05)

                    # Check for external stop (for DIALOG or NAVIGATION)
                    if self._should_stop_externally(priority):
                        logger.error(f"External stop requested for {priority.name}, terminating")
                        process.terminate()
                        try:
                            process.wait(timeout=0.5)
                        except subprocess.TimeoutExpired:
                            process.kill()
                            process.wait()
                        return False

                    # Check for shutdown
                    if self.state == TTSState.SHUTTING_DOWN:
                        logger.error("Shutting down, terminating speech process")
                        process.terminate()
                        try:
                            process.wait(timeout=0.5)
                        except subprocess.TimeoutExpired:
                            process.kill()
                            process.wait()
                        return False

                # Process completed normally
                result = process
            else:
                # For NORMAL priority, just run without monitoring
                result = subprocess.run(
                    cmd,
                    capture_output=True,
                    text=True
                )

            logger.error(f'===> done: "{cmd}" with {priority.name = }')

            if result.returncode == 0:
                return True
            if logger.isEnabledFor(logging.WARNING):
                logger.warning(f"spd-say failed with return code {result.returncode}")
            return False

        except subprocess.TimeoutExpired:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning("TTS command timeout - stopping speech")
            subprocess.run(['spd-say', '-S'], capture_output=True)  # Emergency stop
            return False
        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"TTS execution error: {e}")
            return False

    def _restart_speech_dispatcher(self):
        """Restart speech-dispatcher service with support for multiple commands"""
        try:
            restart_command = self.config['RESTART_LINUX']

            # Separate commands if a && exists
            if '&&' in restart_command:
                commands = [cmd.strip() for cmd in restart_command.split('&&')]
            else:
                commands = [restart_command]

            for cmd in commands:
                cmd_list = shlex.split(cmd)
                result = subprocess.run(
                    cmd_list,
                    timeout=30,
                    capture_output=True
                )

                if result.returncode != 0:
                    if logger.isEnabledFor(logging.WARNING):
                        logger.warning(f"Restart command failed: {cmd}")
                    # We go no to next command even if it fails

            if logger.isEnabledFor(logging.WARNING):
                logger.info("All restart commands executed")
            return True

        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"Failed to restart speech-dispatcher: {e}")

            # Fallback to standard restart methods
            try:
                # Try systemctl first
                subprocess.run(
                    ['systemctl', '--user', 'restart', 'speech-dispatcher'],
                    timeout=30,
                    capture_output=True
                )
                # Then pkill as backup
                subprocess.run(
                    ['pkill', '-f', 'speech-dispatcher'],
                    timeout=10,
                    capture_output=True
                )
                subprocess.run(
                    ['speech-dispatcher', '-d'],
                    timeout=10,
                    capture_output=True
                )
                if logger.isEnabledFor(logging.WARNING):
                    logger.info("Used fallback method to restart speech-dispatcher")
                return True
            except Exception as fallback_error:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error(f"Fallback restart also failed: {fallback_error}")
                return False

    def stop(self):
        """Stop current speech - only used for emergency stops"""
        with self._lock:
            # Clear all external stop flags
            self._clear_external_stop()

            # Stop all speech using -S flag
            subprocess.run(['spd-say', '-S'], capture_output=True)
            # Anti-stutter delay
            time.sleep(self.speech_delay)

    def shutdown(self):
        """Phase 1: Immediate shutdown (non-blocking)"""
        with self._lock:
            self.state = TTSState.SHUTTING_DOWN
            # Linux: Stop all speech immediately
            subprocess.run(['spd-say', '-S'], capture_output=True)

    def wait_for_shutdown(self, timeout=2.0):
        """Phase 2: Wait for complete shutdown (blocking)"""
        with self._lock:
            # Linux: No process to wait for, just cleanup state
            self._current_process = None
            self.state = TTSState.IDLE
            return True

class TTSWindows(TTSBase):
    """Windows TTS implementation using win32com and SAPI"""

    def __init__(self, config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start):
        super().__init__(config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start)
        self.speaker = win32com.client.Dispatch("SAPI.SpVoice")

        # Try to set an English voice
        english_voice = self._get_english_voice()
        if english_voice:
            self.speaker.Voice = english_voice
            if logger.isEnabledFor(logging.INFO):
                logger.info("Windows TTS initialized with English voice")
        else:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning("No English voice found, using system default")

        self.speaker.Volume = self.volume()
        self.current_stream = None
        self._lock = threading.RLock()

    def _get_english_voice(self):
        """Find and return an English voice from available voices"""
        try:
            voices = self.speaker.GetVoices()

            # Priority 1: Look for voices with English indicators in description
            for i in range(voices.Count):
                voice = voices.Item(i)
                voice_description = voice.GetDescription().lower()

                # Check for English language indicators
                english_indicators = [
                    'en-', 'english', ' united states',
                    ' us ', ' uk ', 'british', 'american'
                ]

                if any(indicator in voice_description for indicator in english_indicators):
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(f"Found English voice: {voice_description}")
                    return voice

            # Priority 2: If no specific English voice found, use first available
            if voices.Count > 0:
                first_voice = voices.Item(0)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Using first available voice: {first_voice.GetDescription()}")
                return first_voice

        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"Error finding English voice: {e}")

        return None

    def _execute_speech(self, text, priority=Priority.NORMAL):
        """Execute speech with proper priority handling"""
        with self._lock:
            try:
                # Clear external stop flag for the given priority
                if priority in (Priority.HIGH, Priority.DIALOG, Priority.NAVIGATION):
                    self._clear_external_stop(priority)

                # For ALL priorities, stop any ongoing speech first
                self.stop()

                # Set voice properties
                self.speaker.Volume = int(self.volume())
                self.speaker.Rate = int(self.rate())

                logger.error(f'Windows: executing: "{text[:50]}..." with {priority.name}')

                # HIGH and DIALOG: async execution with interruption checking
                if priority in (Priority.HIGH, Priority.DIALOG):
                    # Using async speech (flags=1) and a waiting loop
                    self.current_stream = self.speaker.Speak(text, 1)  # flags=1 for async
                    self.state = TTSState.SPEAKING

                    while self.current_stream and not self.speaker.WaitUntilDone(50):  # 50ms chunks
                        # Check for external stop (for HIGH and DIALOG)
                        if self._should_stop_externally(priority):
                            self.stop()
                            return False
                        # Check for shutdown
                        if self.state == TTSState.SHUTTING_DOWN:
                            self.stop()
                            return False

                    self.current_stream = None
                    self.state = TTSState.IDLE
                    return True

                # NAVIGATION and NORMAL: async execution without waiting
                self.speaker.Speak(text, 1)  # flags=1 for async
                self.state = TTSState.SPEAKING
                return True

            except Exception as e:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error(f"Windows TTS error: {e}")
                return False

    def stop(self):
        """Stop current speech - optimized for Windows"""
        with self._lock:
            try:
                # Stop immediately
                self.speaker.Speak("", 2)  # flags=2 for immediate stop
                time.sleep(0.1)  # shorter delay
                self.current_stream = None
            except Exception as e:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error(f"Windows TTS stop error: {e}")

    def shutdown(self):
        """Phase 1: Immediate shutdown (non-blocking)"""
        with self._lock:
            self.state = TTSState.SHUTTING_DOWN
            self.stop()

    def wait_for_shutdown(self, timeout=2.0):
        """Phase 2: Wait for complete shutdown (blocking)"""
        with self._lock:
            self.stop()
            return True

class TTSMacOS(TTSBase):
    """macOS TTS implementation with improved interruption handling"""

    def __init__(self, config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start):
        super().__init__(config, volume, rate, pitch, verbosity, speak_volume, speak_volume_start)
        self._current_process = None
        self._lock = threading.RLock()
        self._speech_start_delay = 0.15
        # Faster interruption check interval for better responsiveness
        self._interrupt_check_interval = 0.02  # 20ms instead of 100ms

    def _execute_speech(self, text, priority=Priority.NORMAL):
        """Execute speech on macOS with optimized interruption handling"""
        with self._lock:
            try:
                # ALWAYS stop any current speech first - CRITICAL for macOS
                self._stop_current_speech()

                # Clear external stop flag for the given priority
                self._clear_external_stop(priority)

                # Small delay to avoid race conditions
                time.sleep(0.01)

                cmd = ['say', '-r', self.rate(), text]

                logger.error(f'macOS: executing: "{cmd}" with {priority.name}')

                # Use Popen with stdout/stderr redirection for better performance
                self._current_process = subprocess.Popen(
                    cmd,
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    start_new_session=True  # New session for easier kill
                )
                self.state = TTSState.SPEAKING

                # For NORMAL: wait a bit to let speech start then return
                if priority == Priority.NORMAL:
                    time.sleep(self._speech_start_delay)
                    return True

                # For NAVIGATION, HIGH, DIALOG: wait for completion with interruption checks
                if priority in (Priority.NAVIGATION, Priority.HIGH, Priority.DIALOG):
                    return self._wait_for_completion_with_interrupt(priority)

                return True

            except Exception as e:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error(f"macOS TTS error: {e}")
                return False

    def _stop_current_speech(self):
        """Stop any currently running speech process"""
        if self._current_process:
            try:
                # Try graceful termination first
                if self._current_process.poll() is None:
                    self._current_process.terminate()
                    try:
                        self._current_process.wait(timeout=0.2)
                    except subprocess.TimeoutExpired:
                        # Force kill if not terminated
                        self._current_process.kill()
                        self._current_process.wait()
            except Exception as e:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Error stopping speech process: {e}")
            finally:
                self._current_process = None

        # Clean up any stray say processes
        try:
            subprocess.run(['pkill', '-9', 'say'],
                          stdout=subprocess.DEVNULL,
                          stderr=subprocess.DEVNULL,
                          timeout=0.5)
        except Exception:
            pass

    def _wait_for_completion_with_interrupt(self, priority):
        """Wait for current process to complete with frequent interruption checking"""
        try:
            # Enhanced check for interruption
            while self._current_process and self._current_process.poll() is None:
                # Check for external stop (for DIALOG or NAVIGATION)
                if self._should_stop_externally(priority):
                    logger.error(f"macOS: External stop requested for {priority.name}")
                    self._stop_current_speech()
                    return False
                # Check for shutdown
                if self.state == TTSState.SHUTTING_DOWN:
                    self._stop_current_speech()
                    return False

                # Using a very short sleep for better responsiveness
                # but long enough to not load the PU
                time.sleep(0.01)  # 10ms instead of 20ms

            return self._current_process.returncode == 0 if self._current_process else False

        except Exception as e:
            logger.error(f"Error waiting for speech completion: {e}")
            return False

    def stop(self):
        """Stop current speech"""
        with self._lock:
            self._stop_current_speech()
            time.sleep(self.speech_delay)
            self.state = TTSState.IDLE

    def shutdown(self):
        """Phase 1: Immediate shutdown (non-blocking)"""
        with self._lock:
            self.state = TTSState.SHUTTING_DOWN
            self._stop_current_speech()

    def wait_for_shutdown(self, timeout=2.0):
        """Phase 2: Wait for complete shutdown (blocking)"""
        start_time = time.time()
        with self._lock:
            while (self._current_process and
                   self._current_process.poll() is None and
                   (time.time() - start_time) < timeout):
                time.sleep(0.1)

            self._stop_current_speech()
            self._current_process = None
            self.state = TTSState.IDLE
            return True

class TTSManagerDummy:

    def __init__(self):
        self.enabled = False

    def can_i_use_tts(self, priority=Priority.HIGH):
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug('Speech request rejected: TTS disabled')
        return False

    def queue_speech(self, text, priority=Priority.NORMAL, context=Context.LIMITED, mode=0):
        return

    def stop(self):
        return

    def set_enabled(self, enable):
        return

    def shutdown(self):
        return

    def wait_for_shutdown(self, timeout):
        return

    def stop_dialog_speech(self):
        return

class TTSManager:
    """
    Main TTS manager with priority-based queue and title preservation
    """

    def __init__(
            self, volume, rate, pitch, verbosity, context,
            speak_volume, speak_volume_start, tts_in_config,
            enabled=True):
        self.stop_after_high = False
        self.enabled = enabled
        self.volume = volume
        self.rate = rate
        self.pitch = pitch
        self.verbosity = verbosity
        self.context = context
        self.speak_volume = speak_volume
        self.speak_volume_start = speak_volume_start
        self.config = TTSConfig()
        self.system = platform.system()
        self.available = False
        self.engine = None
        self.tts_in_config = tts_in_config

        # Queue management
        self.high_priority_queue = queue.Queue()
        self.normal_priority_queue = queue.Queue()
        self.pending_title = None
        self._last_spoken_title = None
        self.title_token = self.config.get_title_token()
        self.reset_tokens = self.config.get_reset_tokens()

        # NEW: Track current navigation for self-interruption
        self._current_navigation = None
        self._navigation_lock = threading.RLock()

        # Threading
        self._lock = threading.RLock()
        self._current_request = None
        self._worker_thread = None
        self._shutdown_flag = False

        # Statistics for rapid navigation detection
        self._last_navigation_time = 0
        self._consecutive_requests = 0

        if self.enabled:
            self._initialize_tts()
            self._start_worker()

        # Volume debouncing
        self._pending_volume_request = None
        self._volume_timer = None
        # New lock for volume operations
        self._volume_lock = threading.RLock()

        self._volume_spoken = True
        self._vol_init = datetime.now()

    def _clean_navigation_queue(self):
        """Remove all pending NAVIGATION requests except the last one"""
        with self._navigation_lock:
            if self.high_priority_queue.qsize() == 0:
                return

            # Create a temporary list to hold all requests
            temp_requests = []
            last_navigation = None

            # Extract all requests from queue
            while not self.high_priority_queue.empty():
                try:
                    req = self.high_priority_queue.get_nowait()
                    if req.priority == Priority.NAVIGATION:
                        last_navigation = req  # Keep only the last one
                    else:
                        temp_requests.append(req)  # Keep non-NAVIGATION
                except queue.Empty:
                    break

            # Put back non-NAVIGATION requests
            for req in temp_requests:
                self.high_priority_queue.put(req)

            # Put back the last NAVIGATION request (if any)
            if last_navigation:
                self.high_priority_queue.put(last_navigation)

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Cleaned navigation queue, kept {1 if last_navigation else 0} NAVIGATION requests")

    def stop_navigation_speech(self):
        """Stop currently speaking NAVIGATION priority speech"""
        if self.engine:
            self.engine.request_external_stop(Priority.NAVIGATION)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Requested stop for NAVIGATION speech")

    def _process_pending_title(self):
        """Process pending title after HIGH or NAVIGATION priority completes"""
        if self.pending_title and not self._shutdown_flag:
            # Check if we should actually speak the title
            # Don't speak if another HIGH/NAVIGATION is queued
            if self.high_priority_queue.empty():
                # Add pending title to normal queue
                self.normal_priority_queue.put(TTSRequest(
                    self.pending_title,
                    Priority.NORMAL
                ))
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Queued pending title: {self.pending_title[:50]}...")
            self.pending_title = None

    def can_i_use_tts(self, priority=Priority.NORMAL):
        """ return True if TTS is available and enabled
            and the context is allowed to be spoken
        """
        if self.available and self.enabled:
            if self.tts_in_config():
                return True
            context = self.context()
            if context == 'all':
                return True
            if context == 'window' and \
                    priority.value <= Priority.DIALOG.value:
                return True
            if context == 'limited' and \
                    priority.value <= Priority.HIGH.value:
                return True
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f'Speech request rejected: priority is "{priority.name}" but context is "{context}"')
        return False

    def _initialize_tts(self):
        """Initialize TTS with proper availability checking"""
        try:
            if self.system == "Windows":
                self.available = self._check_windows_availability()
            elif self.system == "Darwin":
                self.available = self._check_macos_availability()
            else:
                self.available = self._check_linux_availability()

            if self.available:
                self.engine = self._create_engine()
                if logger.isEnabledFor(logging.INFO):
                    logger.info(f"TTS initialized successfully for {self.system}")
            else:
                if logger.isEnabledFor(logging.WARNING):
                    logger.warning(f"TTS not available on {self.system}")

        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"TTS initialization failed: {e}")
            self.available = False

    def _create_engine(self):
        """Create the appropriate TTS engine for the current platform"""
        if self.system == "Windows":
            return TTSWindows(self.config, self.volume, self.rate, self.pitch, self.verbosity, self.speak_volume, self.speak_volume_start)
        if self.system == "Darwin":
            return TTSMacOS(self.config, self.volume, self.rate, self.pitch, self.verbosity, self.speak_volume, self.speak_volume_start)
        return TTSLinux(self.config, self.volume, self.rate, self.pitch, self.verbosity, self.speak_volume, self.speak_volume_start)

    def _check_linux_availability(self):
        """Check if spd-say is available on Linux"""
        try:
            # Check if spd-say command exists and works
            result = subprocess.run(
                ['which', 'spd-say'],
                capture_output=True,
                timeout=5
            )
            if result.returncode != 0:
                if logger.isEnabledFor(logging.WARNING):
                    logger.warning("spd-say not found in PATH")
                return False

            # Test that it actually works
            test_result = subprocess.run(
                ['spd-say', '--version'],
                capture_output=True,
                timeout=5
            )
            return test_result.returncode == 0

        except Exception as e:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning(f"Linux TTS check failed: {e}")
            return False

    def _check_windows_availability(self):
        """Check if Windows TTS is available via SAPI"""
        try:
            speaker = win32com.client.Dispatch("SAPI.SpVoice")
            # Test with empty speech
            speaker.Speak("", 1)
            return True
        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"Windows TTS availability check failed: {e}")
            return False

    def _check_macos_availability(self):
        """Check if say command is available on macOS"""
        try:
            result = subprocess.run(
                ['which', 'say'],
                capture_output=True,
                timeout=5
            )
            # say is almost always available on macOS
            return result.returncode == 0

        except Exception as e:
            if logger.isEnabledFor(logging.WARNING):
                logger.warning(f"macOS TTS check failed: {e}")
            return False

    def _start_worker(self):
        """Start the queue processing worker thread"""
        if not self.available or not self.engine:
            return

        self._worker_thread = threading.Thread(
            target=self._process_queues,
            daemon=True,
            name="TTS-Worker"
        )
        self._worker_thread.start()
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("TTS worker thread started")

    def stop_dialog_speech(self):
        """Stop currently speaking DIALOG priority speech"""
        if self.engine:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Requesting stop for DIALOG speech")
            if platform.system() == 'Linux':
                self.stop()
            elif platform.system() == 'Darwin':
                self.engine.request_external_stop(Priority.HIGH)
            else:
                self.engine.request_external_stop()

    def _process_queues(self):
        """Process speech queues with optimized handling for NAVIGATION"""
        while not self._shutdown_flag:
            try:
                # Use shorter timeouts for better responsiveness
                request = None

                # 1. Check for HIGH/DIALOG/NAVIGATION/HELP first (non-blocking)
                try:
                    request = self.high_priority_queue.get_nowait()
                except queue.Empty:
                    pass

                # 2. If no HIGH priority, check NORMAL queue with short timeout
                if not request:
                    try:
                        request = self.normal_priority_queue.get(timeout=0.01)
                    except queue.Empty:
                        time.sleep(0.005)  # Very small sleep for CPU efficiency
                        continue

                # Execute the request if it has actual text
                if request and request.text:
                    self._execute_request(request)

            except Exception as e:
                if logger.isEnabledFor(logging.ERROR):
                    logger.error(f"Queue processing error: {e}")
                time.sleep(0.01)

    def _stop_normal_speech(self):
        """Stop currently speaking NORMAL priority speech"""
        with self._lock:
            # ALWAYS stop NORMAL when HIGH/DIALOG/NAVIGATION/HELP arrives
            if (self._current_request and
                self._current_request.priority == Priority.NORMAL):
                logger.error(f'**** Stopping NORMAL speech for {self.system}')
                self.engine.stop()
                return True

            # ADDITION: Also stop if engine is speaking and we don't know the request
            if (self.engine and
                self.engine.state == TTSState.SPEAKING and
                not self._current_request):
                logger.error(f'**** Stopping unknown speech (assumed NORMAL) for {self.system}')
                self.engine.stop()
                return True

            return False

    def _execute_request(self, request):
        """Execute a TTS request"""
        if not self.engine or self._shutdown_flag:
            return

        with self._lock:
            self._current_request = request
            self.engine.state = TTSState.SPEAKING

        try:
            prefix = ''
            pitch = self.pitch()
            logger.error(f'{pitch = }')
            if pitch != '0':
                if platform.system().lower().startswith('darwin'):
                    prefix = f'[[pbas {pitch}]]'
                elif platform.system().lower().startswith('win'):
                    prefix = f'<pitch absmiddle="{pitch}"/>'
            logger.error(f'{prefix = }')
            transformed_text = tts_transform_to_string([request.text], self.verbosity())
            success = self.engine._execute_speech(
                prefix+transformed_text,
                request.priority
            )
            if not success:
                if logger.isEnabledFor(logging.WARNING):
                    logger.warning(f"Failed to speak: {request.text[:50]}...")
            else:
                if self._is_title(request.text):
                    self._last_spoken_title = request.text
        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"Speech execution error: {e}")
        finally:
            with self._lock:
                # Process pending title after HIGH or NAVIGATION completes
                if request.priority in (Priority.HIGH, Priority.NAVIGATION):
                    self._process_pending_title()

                # Clear current request BEFORE changing state
                self._current_request = None

                if self.engine.state != TTSState.SHUTTING_DOWN:
                    self.engine.state = TTSState.IDLE

        if self.stop_after_high and request.priority == Priority.HIGH:
            self.set_enabled(False)
            self.stop_after_high = False

    def _wait_with_interruption(self, delay):
        """Wait with interruption checks for anti-stutter"""
        chunk_size = 0.01  # 10ms chunks
        remaining = delay

        while remaining > 0 and not self._shutdown_flag:
            time.sleep(min(chunk_size, remaining))
            remaining -= chunk_size

            # Check if we should interrupt
            with self._lock:
                if (self.engine and
                    self.engine.state == TTSState.SHUTTING_DOWN):
                    return False

        return not self._shutdown_flag

    def _should_reset_title(self, text):
        """Check if this message should reset pending titles"""
        return any(reset_token in text for reset_token in self.reset_tokens)

    def _is_title(self, text):
        """Check if text is a title message"""
        return text.startswith(self.title_token)

    def _clean_normal_queue(self):
        """Clean normal priority queue - always clear for new NORMAL"""
        # Drain the queue completely
        while not self.normal_priority_queue.empty():
            try:
                self.normal_priority_queue.get_nowait()
            except queue.Empty:
                break
        if logger.isEnabledFor(logging.DEBUG):
            logger.debug("Cleaned normal priority queue completely")

    def _get_context_value(self):
        context = self.context()
        if context == 'all':
            return Context.ALL.value
        if context == 'window':
            return Context.WINDOW.value
        return Context.LIMITED.value

    def _wake_worker(self):
        """Wake up worker thread for faster processing"""
        # Add an empty request to wake up the worker thread
        try:
            # This will make the worker thread's queue.get() return immediately
            if self.high_priority_queue.qsize() < 10:  # Prevent queue flooding
                self.high_priority_queue.put(TTSRequest("", Priority.NORMAL), block=False)
        except queue.Full:
            pass
        except Exception as e:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Could not wake worker: {e}")

    def _clean_all_queues(self):
        """Clean all queues completely - for rapid navigation scenarios"""
        with self._lock:
            # Clean high priority queue
            while not self.high_priority_queue.empty():
                try:
                    self.high_priority_queue.get_nowait()
                except queue.Empty:
                    break

            # Clean normal priority queue
            while not self.normal_priority_queue.empty():
                try:
                    self.normal_priority_queue.get_nowait()
                except queue.Empty:
                    break

            # Reset state
            self.pending_title = None
            self._current_request = None

            if logger.isEnabledFor(logging.DEBUG):
                logger.debug("Cleaned all queues completely")

    def queue_speech(self, text, priority=Priority.NORMAL, context=Context.LIMITED, mode=0):
        logger.error(f'{text = }')
        logger.error(f'{mode = }')
        if not self.enabled or not self.available or not self.engine:
            return False

        logger.error('config says: {}'.format(self._get_context_value()))
        if context.value > self._get_context_value():
            logger.error('refusing {} = {}'.format(context.name, context.value))
            return False

        if not text or not text.strip():
            return False
        if '%' in text:
            text = text.replace('%', ' percent')
        if priority == Priority.HIGH and ' (error ' in text:
            text = text.split(' (error ')[0]

        # Volume debouncing logic - applies to HIGH and NAVIGATION
        if (priority in (Priority.HIGH, Priority.NAVIGATION) and
            text.startswith(M_STRINGS['volume_set'])):

            with self._volume_lock:
                # Cancel previous timer
                if self._volume_timer:
                    self._volume_timer.cancel()
                    logger.error('\n\nCanceling volume times\n\n')

                # Check if i am allowed to speak volume
                delay_volume_speak = 0
                try:
                    delay_volume_speak = int(self.speak_volume()) / 1000
                except ValueError:
                    pass
                if delay_volume_speak == 0:
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug('Rejecting volume request: tts_speak_volume_change = 0')
                    return True

                # Prepare the TTS notification
                self._pending_volume_request = TTSRequest(text, priority)

                # Check if speaking start start of volume change
                if self.speak_volume_start():
                    now_vol_init = datetime.now()
                    diff = now_vol_init - self._vol_init
                    sec_diff = diff.seconds + diff.microseconds // 1000000
                    logger.error('====================')
                    logger.error(f'{now_vol_init = }')
                    logger.error(f'{self._vol_init = }')
                    logger.error(f'{diff = }')
                    logger.error(f'{sec_diff = }')
                    logger.error(f'{delay_volume_speak = }')
                    logger.error('====================')
                    if sec_diff > 2.5 * delay_volume_speak and \
                            self._volume_spoken:
                        self._vol_init = now_vol_init
                        volume_request = TTSRequest(
                            'Volume change initiated',
                            priority
                        )
                        self.high_priority_queue.put(volume_request)
                        self._volume_spoken = False
                        return True

                # Save volume value
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Holding volume request: {text[:30]} for {delay_volume_speak} sec")

            return True

        # If it's not volume, process any pending volume first
        with self._volume_lock:
            if self._pending_volume_request:
                self._process_pending_volume_immediately()

        if self._is_title(text) and text == self._last_spoken_title:
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Ignoring repeated title: {text[:50]}...")
            return True

        if self._is_title(text):
            self._last_spoken_title = text

        request = TTSRequest(text, priority)

        try:
            if priority in (Priority.HIGH, Priority.DIALOG, Priority.NAVIGATION, Priority.HELP):
                # HIGH/DIALOG/NAVIGATION/HELP priority handling
                if self._should_reset_title(text):
                    self.pending_title = None
                    # Reset if changing title
                    self._last_spoken_title = None
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(f"Reset pending title due to: {text[:50]}...")

                # ALWAYS stop NORMAL when HIGH/DIALOG/NAVIGATION/HELP arrives
                self._stop_normal_speech()

                # NAVIGATION-specific: Self-interruption logic
                if priority == Priority.NAVIGATION:
                    with self._navigation_lock:
                        # Rapid navigation detection
                        current_time = time.time()
                        if current_time - self._last_navigation_time < 0.1:  # 100ms
                            self._consecutive_requests += 1
                            if self._consecutive_requests > 3:
                                self._clean_all_queues()
                                self._consecutive_requests = 0
                        else:
                            self._consecutive_requests = 0

                        self._last_navigation_time = current_time

                        # If current request is also NAVIGATION, stop it
                        if (self._current_request and
                                self._current_request.priority == Priority.NAVIGATION):
                            logger.error('**** stopping previous NAVIGATION')
                            self.engine.request_external_stop(Priority.NAVIGATION)

                        # Clean up any pending NAVIGATION requests
                        self._clean_navigation_queue()

                    # For macOS: Ensure immediate stop before new speech
                    if self.system == "Darwin":
                        self.engine.stop()
                        time.sleep(0.01)  # Small delay to avoid race conditions

                    # Platform-specific: Wake up worker thread for faster processing on non-Linux
                    if self.system != "Linux":
                        self._wake_worker()

                self.high_priority_queue.put(request)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Queued {priority.name} priority: {text[:50]}...")
                return True

            # NORMAL priority request
            # Clean normal queue to avoid accumulation
            self._clean_normal_queue()

            # Stop any currently speaking NORMAL for new NORMAL
            self._stop_normal_speech()

            # Check if HIGH/DIALOG/NAVIGATION/HELP is playing or queued
            if not self.high_priority_queue.empty() or (
                self._current_request and
                self._current_request.priority in (Priority.HIGH, Priority.DIALOG, Priority.NAVIGATION, Priority.HELP)
            ):
                # HIGH/DIALOG/NAVIGATION/HELP is playing or queued
                if self._is_title(text):
                    # Title during HIGH/NAVIGATION/HELP: save as pending
                    self.pending_title = text
                    if logger.isEnabledFor(logging.DEBUG):
                        logger.debug(f"Title saved as pending: {text[:50]}...")
                    return True
                # Non-title NORMAL during HIGH/DLG/NAV/HELP: reject
                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug("Rejected NORMAL request during HIGH/DLG/NAV/HELP playback")
                return False

            # NORMAL priority
            self.normal_priority_queue.put(request)
            if logger.isEnabledFor(logging.DEBUG):
                logger.debug(f"Queued NORMAL: {text[:50]}...")
            return True

        except Exception as e:
            if logger.isEnabledFor(logging.ERROR):
                logger.error(f"Error queuing speech request: {e}")
            return False

    def _process_pending_volume(self):
        """Speak pending volume request after debounce period"""
        with self._volume_lock:
            if self._pending_volume_request:
                volume_request = self._pending_volume_request
                self._pending_volume_request = None
                self._volume_timer = None

                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Processing debounced volume: {volume_request.text[:30]}...")
                # Insert it in the HIGH queue to be spoken
                self.high_priority_queue.put(volume_request)
                self._volume_spoken = True

    def _process_pending_volume_immediately(self):
        """Forced speech for pending volume SR before any other request"""
        with self._volume_lock:
            if self._pending_volume_request:
                volume_request = self._pending_volume_request
                self._pending_volume_request = None
                if self._volume_timer:
                    self._volume_timer.cancel()
                    self._volume_timer = None

                if logger.isEnabledFor(logging.DEBUG):
                    logger.debug(f"Processing pending volume immediately: {volume_request.text[:30]}...")
                self.high_priority_queue.put(volume_request)
                self._volume_spoken = True

    def stop(self):
        """Stop current speech"""
        if self.engine:
            self.engine.stop()

    def shutdown(self):
        """Phase 1: Immediate shutdown (non-blocking)"""
        self._shutdown_flag = True
        if self.engine:
            self.engine.shutdown()

    def wait_for_shutdown(self, timeout=2.0):
        """Phase 2: Wait for complete shutdown (blocking)"""
        if self._worker_thread and self._worker_thread.is_alive():
            self._worker_thread.join(timeout=timeout)

        if self.engine:
            return self.engine.wait_for_shutdown(timeout)
        return True

    def set_enabled(self, enabled):
        """Enable/disable TTS globally"""
        old_state = self.enabled
        self.enabled = enabled

        if enabled and not old_state:
            # Completely reinitialize TTS
            self.shutdown()
            self.wait_for_shutdown()

            # Clear queues for restart
            while not self.high_priority_queue.empty():
                try:
                    self.high_priority_queue.get_nowait()
                except queue.Empty:
                    break
            while not self.normal_priority_queue.empty():
                try:
                    self.normal_priority_queue.get_nowait()
                except queue.Empty:
                    break

            # Reset all state
            self._shutdown_flag = False
            self.engine = None
            self.available = False

            self._initialize_tts()
            self._start_worker()
            logger.error(f'TTS reinitialized - {self.engine = }')

        elif not enabled and old_state:
            # Turning OFF - shutdown gracefully
            self.shutdown()
            self.wait_for_shutdown()
            self.engine = None

    def is_available(self):
        """Check if TTS is available and enabled"""
        return self.enabled and self.available

    def get_status(self):
        """Get detailed status information"""
        engine_state = self.engine.state.name if self.engine else 'NO_ENGINE'
        return {
            'system': self.system,
            'enabled': self.enabled,
            'available': self.available,
            'engine_initialized': self.engine is not None,
            'state': engine_state,
            'high_queue_size': self.high_priority_queue.qsize(),
            'normal_queue_size': self.normal_priority_queue.qsize(),
            'pending_title': self.pending_title is not None
        }

